// $Id: audsld.cc,v 1.16 2006/10/17 09:41:09 gson Exp $ 
//
// User-mode AuDSL modem daemon main program.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#include <stdio.h>
#include <iostream>
#include <fcntl.h>
#include <err.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#ifdef LINUX_TUNTAP
#include <linux/if.h>
#include <linux/if_tun.h>
#endif

#include "open-audio.h"
#include "usermode-modem.h"

#define PATH_AUDSLD_PID "/var/run/audsld.pid"
#define PATH_AUDSLD_SHMID "/var/run/audsld.shmid"

char *progname;

int shmid = 0;	/* SYSV shared memory segment ID */

void error(const char *s)
{
#if 0
    syslog(LOG_ERR, "%s: %s: %m", progname, s);
#else
    fprintf(stderr, "%s: %s: %s\n", progname, s, strerror(errno));
#endif
    exit(1);
}

// Catch SIGINT and call exit so that profiling information
// gets written to disk before we die.  Also clean up the
// PID file and the SYSV shared memory stuff.

extern "C" void sig_handler(int) {
    (void) unlink(PATH_AUDSLD_PID);
#ifdef SHM_INSTRUMENT
    (void) unlink(PATH_AUDSLD_SHMID);
    if (shmid)
	(void) shmctl(shmid, IPC_RMID, 0);
#endif
    exit(1);
}

void shell_command(char **argv)
{
    switch (fork()) {
    case -1: {
	error("fork failed");
	break;
    }
    case 0: {
        // child
	execv(argv[0], argv);
	error("could not execute shell command");
	break;
    }
    default: {
	int status;
	wait(&status);
	if (status) {
	    error("shell command failed");
	}
    }
    }
}

int main(int argc, char **argv)
{
    progname = argv[0];
    argc--, argv++;

    int use_shm = 0;	// If true, use shared memory
    int foreground = 0;

    char *dev_audio = DEFAULT_AUDIO_DEV;
    char *dev_tun = DEFAULT_TUN_DEV;

    while (argc) {
	if (!strcmp(argv[0], "-s")) {
	    use_shm = 1;
	    argc--, argv++;
	} else if (!strcmp(argv[0], "-a") && (argv[0][1] || argc > 1)) {
	    dev_audio = argv[0][2] ? (argv[0] + 2) : (argc--, argv++, argv[0]);
	    argc--, argv++;
	} else if (!strcmp(argv[0], "-t") && (argv[0][1] || argc > 1)) {
	    dev_tun = argv[0][2] ? (argv[0] + 2) : (argc--, argv++, argv[0]);
	    argc--, argv++; 
	} else if (!strcmp(argv[0], "-f")) {
	    foreground = 1;
	    argc--, argv++;
	} else {
	    fprintf(stderr, "usage: %s [-s] [-a /dev/audioN] [-t /dev/tunN]\n",
		    progname);
	    exit(1);
	}
    }

#ifdef NEED_AUCONF_SH
    char *auconf_cmd[] = { "/bin/sh", "auconf.sh", dev_audio, 0 };
    shell_command(auconf_cmd);
#endif

#if defined(NEED_TUNCONF_SH)
#if !defined(LINUX_TUNTAP) // still done, but later
    char *tunconf_cmd[] = { "/bin/sh", "tunconf.sh", dev_tun, 0 };
    shell_command(tunconf_cmd);
#endif
#endif

    int tun_fd = open(dev_tun, O_RDWR);
    if (tun_fd < 0)
	error("tun open");
    int r = fcntl(tun_fd, F_SETFL, fcntl(tun_fd, F_GETFL, 0) | O_NONBLOCK);
    if (r < 0)
	error("tun fcntl");

#ifdef LINUX_TUNTAP
    struct ifreq ifr;
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TUN; 
    r = ioctl(tun_fd, TUNSETIFF, (void *) &ifr);
    if (r < 0)
	error("tun ioctl(TUNSETIFF)");
    char if_tun[100];
    strncpy(if_tun, ifr.ifr_name, sizeof if_tun);
    
    // On Linux, the tun network interface does not exist until we
    // have opened /dev/net/tun and done the TUNSETIFF IOCTL,
    // so we must do that before running tunconf.sh.
    char *tunconf_cmd[4];
    tunconf_cmd[0] = "/bin/sh";
    tunconf_cmd[1] = "tunconf.sh";
    tunconf_cmd[2] = if_tun;
    tunconf_cmd[3] = 0;
    shell_command(tunconf_cmd);
#endif

    setpriority(PRIO_PROCESS, 0, -20);
    
    int audio_fd = open_audio(dev_audio);

#ifdef SHM_INSTRUMENT
    void *addr;
    if (use_shm) {
	// Create the modem object in a shared memory segment so that the
	// status monitor can access it.
	shmid = shmget(IPC_PRIVATE, sizeof(UserModeAudslModem), IPC_CREAT|0777);
	if (shmid == -1)
	    error("shmget");
	addr = shmat(shmid, (char *) 0, 0);
	if (addr == ((void *) -1))
	    error("shmat");
	FILE *f = fopen(PATH_AUDSLD_SHMID, "w");
	if (! f)
	    error("opening shmid file");
	fprintf(f, "%08x\n", shmid);
	fclose(f);
    } else {
	addr = malloc(sizeof(UserModeAudslModem));
    }
    UserModeAudslModem &um = 
	*new (addr) UserModeAudslModem(tun_fd, audio_fd);
#else
    UserModeAudslModem um(tun_fd, audio_fd);
#endif

    if (! foreground)
	daemon(0, 0);

    FILE *f = fopen(PATH_AUDSLD_PID, "w");
    if (! f)
	error("opening pid file");
    fprintf(f, "%d\n", getpid());
    fclose(f);
    
    signal(SIGINT, sig_handler);
    signal(SIGTERM, sig_handler);
    
    struct timeval next;
    gettimeofday(&next, 0);
    struct timeval delta;
    delta.tv_sec = 0; delta.tv_usec = 10000;

    for (;;) {

	um.modem.tick();

#ifdef AUTO_TESTING
	if (um.modem.receiver.state == RX_ERROR &&
	    um.modem.receiver.nsamp > 10000)
	{
	    printf("%f\n", sqrt(um.modem.receiver.mse));
	    exit(0);
	}
#endif
	
	timeradd(&next, &delta, &next);

	struct timeval now, left;
	gettimeofday(&now, 0);
	if (timercmp(&now, &next, >=)) {
	    // no time to sleep
	} else {
	    // sleep for the remainder of the cycle
	    timersub(&next, &now, &left);
	    select(0, 0, 0, 0, &left);
	}
    }
}


