// $Id: bitfifo.h,v 1.7 1999/01/28 21:33:15 gson Exp $
//
// FIFO buffer for bit streams.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

//
// BitFIFO<N>: A FIFO buffer of bits.
//
// N is the FIFO size in bits; it must be a multiple of the word size.
//
// w_word is the index of the word currently being filled, with 
// (w_bit = 0 to 31 bits) already used at the least significant end, 
// and (32-w_bit = 1 to 32) free bits at the most significant end.  
//
// The free bits are always initialized to zero (?)
//

// Determine the number of bits per word.
#define BPW ((int) sizeof(int) * 8)

template <int N>
class BitFIFO {
public:
    BitFIFO() 
	: w_word(0), w_bit(0), r_word(0), r_bit(0)
    {
	assert(N % BPW == 0);
	buf[0] = 0;
    }
	
    void push(int bits, int nbits) {
	while (nbits) {
	    int chunkbits = BPW - w_bit;
	    if (chunkbits > nbits) chunkbits = nbits;
	    buf[w_word] |= bits << w_bit;
	    bits >>= chunkbits;
	    nbits -= chunkbits;
	    w_bit += chunkbits;
	    if (w_bit == BPW) {
		++w_word; if (w_word >= N/BPW) w_word -= N/BPW;
		assert(w_word != r_word);
		buf[w_word] = 0;
		w_bit = 0;
	    }
	}
    }
    
    int pop(int nbits = 1) {
	int pos = 0;
	int result = 0;
	assert(size() >= nbits);
	while (nbits) {
	    int chunkbits = BPW - r_bit;
	    if (chunkbits > nbits) chunkbits = nbits;
	    int tmp = (buf[r_word] >> r_bit);
	    if (chunkbits < BPW)
		tmp &= ((1 << chunkbits) - 1);
	    result |= tmp << pos;
	    nbits -= chunkbits;
	    r_bit += chunkbits;
	    pos += chunkbits;
	    assert (r_bit <= BPW);
	    if (r_bit == BPW) {
		++r_word; if (r_word >= N/BPW) r_word -= N/BPW;
		r_bit = 0;
	    }
	}
	return result;
    }

    int size() {
	int s = (w_bit + w_word * BPW) - (r_bit + r_word * BPW);
	if (s < 0)
	    s += N;
	return s;
    }
private:
    
    int w_word;
    int w_bit;

    int r_word;
    int r_bit;

    unsigned int buf[N/BPW];
};
