// $Id: dsp.h,v 1.40 2006/08/17 15:01:54 gson Exp $
//
// DSP utilities.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#ifndef _DSP_H
#define _DSP_H

#include <assert.h>
#include <math.h>
#include <string.h> // memcpy

// Functions

static inline double sqr(double x) { return x * x; }
static inline float sqr(float x) { return x * x; }

static inline double sinc(double x) {
    if (x == 0)
	return 1.0;
    x *= M_PI;
    return sin(x) / x;
}

// Basic DSP classes

template <class T>
class Impulse {
public:
    Impulse()
	: n(0) { }
    T operator()() { return (n++ == 0) ? 1 : 0; }
private:
    int n;
};

template <class T, int L>
class SimpleShiftReg {
public:
    SimpleShiftReg() {
	i = 0;
	for (int j = 0; j < L; j++)
	    a[j] = 0;
    }
    // Access the value whose age is "index" samples.
    // (*this)[0] is the value that was last pushed.
    T &operator[](int index) {
	assert(index >= 0 && index < L);
	int ai = i - index;
	if (ai < 0) ai += L;
	return a[ai];
    }
    T push(T v) {
	i++; if (i >= L) i -= L;
	T r = a[i];
	a[i] = v;
	return r;
    }
 private:
    T a[L];
    // Position of value last pushed.  Wraps modulo L.
    int i;
};

template <class T, int L>
class Delay: private SimpleShiftReg<T, L> {
public:
    T operator()(T v) {
	return push(v);
    }
};


// A ContiguousSignalBuffer object is a shift register of 
// values of type T.  It offers efficient access
// to the last L values pushed, with the special
// feature that these last L values are guaranteed
// to be contiguous in memory, newest value first.
// The parameter LL determines the internal buffer
// size.  For efficiency, it should be large compared to L.

template <class T, int L, int LL>
class ContiguousSignalBuffer {
public:
    ContiguousSignalBuffer() {
	assert(L <= LL);
	reset();
    }
    void reset() {
	for (int j = 0; j < LL+L-1; j++)
	    a[j] = 0;
	i = LL;
    }
    void push(T v) {
	--i;
	if (i < 0) {
	    memcpy(&a[LL], &a[0], (L-1)*sizeof(a[i]));
	    i = LL-1;
	}
	a[i] = v;
    }
    // Access the value whose age is "index" samples.
    // (*this)[0] is the value that was last pushed.
    const T &operator[](int index) const {
	assert(index >= 0 && index < L);
	return a[i+index];
    }
    T &operator[](int index) {
	assert(index >= 0 && index < L);
	return a[i+index];
    }
    int length() const { return L; }    
 private:
    T a[LL+L-1];
    int i;  // Position of value last pushed.
};

// A first-in, first-out buffer.

template <class T, int L>
class FIFO {
public:
    FIFO()
	: ip(0), op(0) { }
    void push(T v) { a[ip] = v; ++ip; if (ip >= L) ip -= L; }
    T operator()() { T r = a[op]; ++op; if (op >= L) op -= L; return r; }
    bool empty() { return ip == op; }
    int size() { int s = ip - op; if (s < 0) s += L; return s; }
private:
    T a[L];
    int ip, op; // Input pointer, output pointer
};

template <class S0, class S1, class T>
static inline
T convolve(const S0 &s0, const S1 &s1, T init, int n) {
    T sum = init;
    for (int j = 0; j < n; j++)
	sum += s0[j] * s1[j];
    return sum;
}


// A FIR filter for signals of type T, with weights of type WT,
// of length N implemented in terms of a ContinuousSignalBuffer 
// with buffer length NN, operating on samples of type T.

template <class T, class WT, int N, int NN>
class CFIR: public ContiguousSignalBuffer<T, N, NN> {
public:
    CFIR() { }
    // Construct filter with given taps.
    CFIR(const WT *w_in) {
	for (int i = 0; i < N; i++) {
	    w[i] = w_in[i];
	}
    }
    T convolve() { return ::convolve(*this, w, T(0.0f), N); }
    T operator()(T in) {
	push(in);
	return convolve();
    }
    WT w[N];
};

// A CFIR whose impulse response is defined by a filter kernel
// function object.

template <class T, class WT, class Kernel, int N, int NN>
class KCFIR: public CFIR<T, WT, N, NN> {
public:
    // Construct filter with an impulse  response defined by the
    // filter kernel function k (of type Kernel).
    KCFIR(const Kernel &k) {
	int n2 = N/2;
	for (int i = -n2; i < n2; i++) {
	    this->w[n2+i] = k(i);
	}
    }
};

static inline float conj(float f) { return f; }

// A least-mean-square adaptive FIR filter.

template <class T, class WT, int L, int LL>
class LMSFIR: public CFIR<T, WT, L, LL> {
public:
    LMSFIR() { reset(); }
    void reset() {
	for (int j = 0; j < L; j++)
	    this->w[j] = 0;
    }
    // Adjust coefficients according to error signal "e".
    void adjust(WT e) {
	for (int j = 0; j < L; j++)
	    this->w[j] -= e * conj(this->operator[](j));
    }
};


#endif
