// $Id: echo.h,v 1.47 1999/01/28 21:33:16 gson Exp $
//
// Echo canceller
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#include <stdio.h>
#include <math.h>

#include "protocol.h"
#include "pam-mapper.h"
#include "dsp.h"
#include "echo-buffer.h"
#include "lfsr.h"

//#define PLOT_EC_RES	// Plot the echo canceller residual
#ifdef PLOT_EC_RES
#include "plot-signal.h"
#endif

enum EchoCancellerState {
    ECHO_STATE_INITIAL,
    ECHO_STATE_PRIME,   // Send a short sounding burst
    ECHO_STATE_SEARCH,  // Search for echo of sounding burst
    ECHO_STATE_TRAIN,   // Train the adaptive filter
    ECHO_STATE_WAIT,    // Wait for peer to do his own training
    ECHO_STATE_STEADY   // Do the actual cancelling
};

#define ECHO_NEXT(s) (state = s, nsamp = 0)

// BUFLEN is the length of the buffer for transmitted data; it 
// determines the maximum output-to-input delay the canceller
// is able to tolerate. 
//
// Only part of the delayed output signal actually participates
// in cancellation; FIRLEN is the length of this part.

template <class T, int TXBUFLEN, int FIRLEN,
	  class LINEIN, class LINEOUT, class TXIN, class RXOUT>
class EchoCanceller {
public:
    // History buffer for received data.  This is used
    // only in the priming / search stage, for correlating
    // received data with the priming burst.
    typedef ContiguousSignalBuffer<T, FIRLEN, 200> rxbuf_t;
    rxbuf_t rxbuf;

    // History buffer for transmitted data.
    EchoBuffer<T, TXBUFLEN, FIRLEN> txbuf;
    
    EchoCancellerState state;
    int nsamp;
    float max_correlation;
    int offset;
    int wait;
    float coefs[FIRLEN];

    LINEIN &linein;
    LINEOUT &lineout;
    TXIN &txin;
    RXOUT &rxout;

    LFSR<1> lfsr;
    PAMMapper<1> mapper;

    EchoCanceller(LINEIN &lin, LINEOUT &lout, TXIN &tin, RXOUT &rout) 
	: linein(lin), lineout(lout),
	  txin(tin), rxout(rout) { reset(); }

    void reset() {
	state = ECHO_STATE_INITIAL;
	nsamp = 0;
	max_correlation = 0;
	wait = 2000;
	txbuf.reset();
	for (int i = 0; i < FIRLEN; i++)
	    coefs[i] = 0;
    }
    
    void send_training() {
	float txsignal =
	    nsamp < EC_TRAIN_TIME / 2 ?
	    mapper(lfsr()) * 0.5 :
	    float(txin());

	lineout.push(txsignal);
	txbuf.push(txsignal);
    }
    
    void run() {
	float inp = tofloat(linein());

	switch (state) {
	case ECHO_STATE_INITIAL: {
	    ECHO_NEXT(ECHO_STATE_PRIME);
	    txin.training_mode(true);
	    // fallthrough
	}
	    
	case ECHO_STATE_PRIME: {
	    rxbuf.push(inp);
	    send_training();
	    if (nsamp == FIRLEN) {
		ECHO_NEXT(ECHO_STATE_SEARCH);
	    }
	    break;
	}
	    
	case ECHO_STATE_SEARCH: {
	    EchoSubarray<T, TXBUFLEN, FIRLEN> subarray(txbuf, 0);
	    rxbuf.push(inp);
	    T sum = 0;
	    for (int i = 0; i < FIRLEN; i++) {
		sum += rxbuf[i] * subarray[i];
	    }
	    float correlation = sum * sum;
	    if (correlation > max_correlation) {
		max_correlation = correlation;
		offset = nsamp;
	    }
	    
	    lineout.push(0.0f);
	    
	    if (nsamp == TXBUFLEN - FIRLEN - 2) {
		ECHO_NEXT(ECHO_STATE_TRAIN);
	    }
	    break;
	}
	    
	case ECHO_STATE_TRAIN: {
	    EchoSubarray<T, TXBUFLEN, FIRLEN>
		subarray(txbuf, offset - FIRLEN/2);
	    float rxsignal = inp;
	    float e = rxsignal - ::convolve(subarray, coefs, T(0), FIRLEN);
	    for (int j = 0; j < FIRLEN; j++) {
		coefs[j] += e * subarray[j] * 0.04;
	    }

#ifdef PLOT_EC_RES
	    static SignalPlot plot("ecres");
	    plot.push(to16(rxsignal));
	    plot.push(to16(e));
#endif	
	    send_training();

	    if (nsamp == EC_TRAIN_TIME) {
		txin.training_mode(false);
		ECHO_NEXT(ECHO_STATE_WAIT);
	    }
	    break;
	}

	case ECHO_STATE_WAIT: {
	    if (nsamp == wait)
		ECHO_NEXT(ECHO_STATE_STEADY);
	    float txsignal = 0;
	    lineout.push(txsignal);
	    txbuf.push(txsignal);
	    break;
	}
	    
	case ECHO_STATE_STEADY: {
	    float txsignal = txin();
	    EchoSubarray<T, TXBUFLEN, FIRLEN>
		subarray(txbuf, offset - FIRLEN/2);
	    float echo_est = ::convolve(subarray, coefs, T(0), FIRLEN);
	    float signal = inp - echo_est;
#if 0
	    if (nsamp > 100000 && ((nsamp & 0x0F) == 0))
		for (int j = 0; j < FIRLEN; j++)
		    coefs[j] += signal * subarray[j] * 0.00001;
#endif
	    rxout.push(signal);
	    lineout.push(txsignal);
	    txbuf.push(txsignal);
	    break;
	}
	}
	nsamp++;
    }

    int fir_length() { return FIRLEN; }
};
