// $Id: fixed-point.h,v 1.10 1999/01/28 21:33:16 gson Exp $
//
// Support for fixed-point computation.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#ifndef _FIXED_POINT_H
#define _FIXED_POINT_H

extern int n_saturations;

void saturated();

// s16: signed 16-bit fixed-point signal data

class s16;
inline s16 i2s16(int i);

class s16 {
public:
    s16() { }
    explicit s16(int i)
	: v(i) { }

    // Conversions
    inline s16(float f);
    inline operator float();

    s16 &operator += (s16 b) { return *this = i2s16(v + b.v); }
    s16 &operator -= (s16 b) { return *this = i2s16(v - b.v); }    
    
    short v;
};

inline s16 
i2s16(int i) {
    if (i < -32767) {
#ifndef NDEBUG
	saturated();
#endif
	i = -32767;
    }
    if (i > 32767) {
#ifndef NDEBUG
	saturated();
#endif
	i = 32767;
    }
    s16 r;
    r.v = i;
    return r;
}

// Convert a floating-point sample into a 16-bit signed integer sample.

inline s16 to16(float f) {
    f *= 32767.0f;
    f += 32768.5f;
    return i2s16 (((int) f) - 32768);
}

// Convert a 16-bit signed integer sample into a floating-point sample.

inline float tofloat(s16 s) {
    return (float)(1.0 / 32767) * s.v;
}

s16::s16(float f) { *this = to16(f); }
s16::operator float() { return tofloat(*this); }

// Dummy conversions
inline float tofloat(float s) { return s; }
inline s16 to16(s16 s) { return s; }

// Arithmetic
inline s16 operator +(s16 a, s16 b) { return i2s16(a.v +  b.v); }
inline s16 operator -(s16 a, s16 b) { return i2s16(a.v -  b.v); }
inline s16 operator *(s16 a, s16 b) { return i2s16((a.v * b.v) >> 16); }

// An int-sized unsgned fixed point number with an I-bit integer part.  
// The rest of the bits form the fraction.

#define INT_BITS 32 // XXX

#define POW2D(i) ((1 << ((i) & 31)) * (((i) & 32) ? 4294967296.0 : 1.0))
#define POW2F(i) ((float) POW2D(i))

template <int I>
class UFixed {
public:
    // The compiler-generated copy constructior and assignment are OK.
    UFixed() { v = 0; }
    UFixed(double f) : v((unsigned int) (f * POW2D(INT_BITS-I))) { }
    UFixed(float f) : v((unsigned int) (f * POW2F(INT_BITS-I))) { }    
    operator double() const { return (1.0 / POW2D(INT_BITS-I)) * v; }
    operator float() const { return (1.0f / POW2F(INT_BITS-I)) * v; }

    static UFixed frombits(int i) { UFixed f; f.v = i; return f; }
    unsigned int bits() const { return v; }

    friend UFixed<I> operator +(const UFixed<I> &a, const UFixed<I> &b)
        { return frombits(a.v + b.v); }
    friend UFixed<I> operator - (const UFixed<I> &a, const UFixed<I> &b)
        { return frombits(a.v - b.v); }
    friend bool operator >=(const UFixed<I> &a, const UFixed<I> &b)
       { return a.v >=  b.v; }

    UFixed<I> & operator += (const UFixed<I> &b) { v += b.v; return *this; }
    UFixed<I> & operator -= (const UFixed<I> &b) { v -= b.v; return *this; }

private:
    unsigned int v;
};

#endif
