// $Id: handshake.cc,v 1.26 2002/08/18 02:40:28 gson Exp $
//
// The modem handshake protocol state machine. 
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#include "handshake.h"
#include "fixed-point.h"

// #define PLOT_HS_INPUT
#ifdef PLOT_HS_INPUT
#include "plot-signal.h"
#endif

#define NEXT(s) (state = s, nsamp = 0)

// Frequencies for pilot tones sent by each side during the
// handshake.  Chosen to be incommensurate, to avoid the 
// pilot tone detectors locking onto harmonics.

#define FQA 0.67957046	// Center frequency for side A
#define FQB 0.25	// Center frequency for side B
#define FQD 0.125 	// Half bandwidth of pilot detector filter

HandshakeStateMachine::HandshakeStateMachine(Gain &rec_gain)
    : filter_a(GaussianWindowedBandpassKernel(FQA-FQD, FQA+FQD, 32, 2.5)),
      filter_b(GaussianWindowedBandpassKernel(FQB-FQD, FQB+FQD, 32, 2.5)),
      pll_a(FQA * M_PI),
      pll_b(FQB * M_PI),
      osc_a(FQA * M_PI),
      osc_b(FQB * M_PI),
      gain(rec_gain)
{
    reset();
}

void HandshakeStateMachine::reset() 
{
    identity = ID_X;
    state = HS_STATE_TRY_B;
    pll_a.reset(); pll_b.reset();
    osc_a.reset(); osc_b.reset();
    rng = 0;
    nsamp = 0;
    nlisten = 48000;
    inmin = inmax = 0;
    bestgain = -1;
}

float HandshakeStateMachine::run(float input)
{
    float r = 0;

#ifdef PLOT_HS_INPUT
	static SignalPlot plot("hsin");
	plot.push(input);
#endif

    rng ^= rng << 1; rng ^= to16(input).v;

    switch (state) {

    case HS_STATE_TRY_B: {
	// We optimize for the case where the modems are not
	// turned on simultaneously.  When a modem has been
	// trying to handhake for a long time, it will mostly
	// be trying to be A.  Thus, a new modem coming on-line
	// should begin by trying to be B and listening for A.
	
	r = osc_b();
	pll_a(filter_a(input));
	if (pll_a.locked())
	    NEXT(HS_STATE_HEARD_A);
	if (nsamp == 48000) {
	    nlisten = 5*48000 + (rng & 0x7FFFF);
	    NEXT(HS_STATE_DRAIN_B);
	}
	break;
    }

    case HS_STATE_DRAIN_B: {
	r = 0;
	if (nsamp == 20000)
	    NEXT(HS_STATE_TRY_A);
	break;
    }

    case HS_STATE_TRY_A: {
	// This is the state we will be in most of the time, waiting
	// for a new modem to come on line at the other end (or to
	// try its occasional B role).
	r = osc_a();
	pll_b(filter_b(input));
	if (pll_b.locked())
	    NEXT(HS_STATE_HEARD_B);
	if (nsamp == nlisten)
	    NEXT(HS_STATE_DRAIN_A);
	break;
    }

    case HS_STATE_DRAIN_A: {
	r = 0;
	if (nsamp == 20000)
	    NEXT(HS_STATE_TRY_B);
	break;
    }

    case HS_STATE_HEARD_A: {
	r = osc_b();
	if (nsamp == 48000) {
	    identity = ID_B;
	    NEXT(HS_STATE_PRE_AGC_WAIT);
	}
	break;
    }

    case HS_STATE_HEARD_B: {
	r = osc_a();
	if (nsamp == 48000) {
	    identity = ID_A;
	    NEXT(HS_STATE_PRE_AGC_WAIT);
	}
	break;
    }

    case HS_STATE_PRE_AGC_WAIT: {
	if (nsamp == 24000)
	    NEXT(HS_STATE_AGC);
	break;
    }

    case HS_STATE_AGC: {
	// XXX should use actual transmit signal
	r = lfsr() ? 1.0f : -1.0f;

	int t = nsamp - 24000;
	if (t >= 0) {
	    if (t % 1000 == 0) {
		gain.set_gain(t / 1000);
		inmin = inmax = 0;
	    }
	    if (t % 1000 >= 500) {
		if (input > inmax) inmax = input;
		if (input < inmin) inmin = input;
	    }
	    if (t % 1000 == 999) {
		if (inmin > -0.5 && inmax < 0.5)
		    bestgain = t / 1000;
	    }
	}
	if (t == 48000) {
	    assert(bestgain >= 0);
	    gain.set_gain(bestgain);
	    NEXT(HS_STATE_SILENT);
	}
	break;
    }

    case HS_STATE_SILENT: {
	int nsilent = 24000;
	if (identity == ID_B)
	    nsilent += EC_SILENCE_TIME;
	if (nsamp == nsilent)
	    NEXT(HS_STATE_END);
	break;
    }

    case HS_STATE_END: {
	break;
    }
    }

    nsamp++;
    return r;
}

