// $Id: hdlc.h,v 1.9 1999/01/28 21:33:18 gson Exp $
//
// HDLC framing / deframing, with bit stuffing.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

// A HDLC framer.  

// RFC1662 says:
// "All octets are transmitted least significant bit first."

template <class I, class O, class P>
class HDLCFramer {
public:

    // Construct a framer that reads packets from "in" and writes bits to
    // "out".

    // The input object "in" (of class I) should have methods empty(), 
    // front(), and pop(), like an STL queue.  Its elements (of class P) 
    // represent the packets to transmit, and should have the methods
    // data() and length(), like STL strings.

    // The output object "out" (of class O) should have the method
    // push(int bits, int nbits).  When multiple bits are pushed at
    // once, the least significant bit is considered the first one.
    
    HDLCFramer(I &in, O &out)
	: input(in), output(out), packet(0), bytes_left(0), ones(0)
    {
	output_flag();
    }

    // Run the framer to generate at least "minbits" output bits.
    // May generate up to minbits+15 bits, though. 

    void run(int minbits) {
	int nbytes = ((minbits + 7) >> 3);
	while (nbytes--) {
	    if (input.empty()) {
		output_flag();
		continue;
	    } 
	    
	    if (! bytes_left) {
		const P &p = input.front();
		packet = p.data();
		bytes_left = p.length();
		ones = 0;
	    }

	    int byte = *packet++;
	    
	    for (int i = 0; i < 8; i++) {
		int bit = byte & 0x01;
		// Do bit stuffing.
		if (bit)
		    ones++;
		else
		    ones = 0;
		output.push(bit, 1);
		// If this was the fifth consecutive "1",
		// insert a "0" after it.
		if (ones == 5) {
		    output.push(0, 1);
		    ones = 0;
		}
		byte >>= 1;
	    }

	    bytes_left--;

	    if (! bytes_left) {
		output_flag();
		input.pop();
	    }
	}
    }
    void output_flag() {
	output.push(0x7E, 8);
	ones = 0;
    }
private:
    I &input;
    O &output;
    const char *packet;
    int bytes_left;
    int ones;
};

// A HDLC deframer.

template <class O>
class HDLCDeframer {
public:
    HDLCDeframer(O &o)
	: output(o), r(0), byte(0), nbits(0) { }

    void run(int bit) {
	r <<= 1; r |= bit;
	if ((r & 0x7F) == 0x7E) {
	    // ...1111110: must be a flag or break ending here
	    output.push_flag();
	    nbits = 0;
	} else if ((r & 0x3F) == 0x3E) {
	    // ...111110: must be bit stuffing, ignore this bit
	} else {
	    // data; shift into byte register from most significant end
	    byte >>= 1; byte |= (bit << 7);
	    ++nbits;
	    if (nbits == 8) {
		output.push_bits(byte, nbits);
		nbits = 0;
	    }
	}
    }
    
private:
    O &output;
    unsigned int r; 	// big-endian shift register for flag detection
    int byte;	// byte buffer
    int nbits;  // number of valid bits in the byte buffer
};
