// $Id: modem.h,v 1.52 2006/10/18 13:07:53 gson Exp $
//
// Generic modem class.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#ifndef _MODEM_H
#define _MODEM_H

//#define QAM

#include <string.h>

typedef float sample_t;

//#include "fixed-point.h"
//typedef s16 sample_t;

#ifdef QAM
#include "qam-rx.h"
#define Receiver QAMReceiver
#include "qam-tx.h"
#define Transmitter QAMTransmitter
#define RX_ERROR QAM_RX_STATE_ERROR
#else
#include "pam-rx.h"
#define Receiver PRPAMReceiver
#include "pam-tx.h"
#define Transmitter PRPAMTransmitter
#define RX_ERROR PRPAM_RX_STATE_ERROR
#endif

#include "hdlc.h"
#include "bitfifo.h"
#include "echo.h"
#include "handshake.h"

#define MTUMAX 1500	// Maximum maximum (sic) transmission unit

#define CHUNKSIZE 1000 	// Number of samples to process in one go.

#define N_EC_TAPS 32	// Number of echo canceller taps
#define N_EQ_TAPS 16	// Number of equalizer taps

// An object for turning a bit stream into packets.
// XXX silly restrictions

template <class PacketWriter>
class BitStreamPacketizer {
public:
    BitStreamPacketizer(PacketWriter &packet_writer)
	: pw(packet_writer) { reset(); }
    void reset() {
	n = 0;
	n_bits = 0;
	n_errs = 0;
	n_pkts = 0;
    }
    void push_bits(int d, int nbits) {
	if (nbits != 8 || n == sizeof(buf)) {
	    n = -1; // this packet is in error
	    n_errs++;
	} else {
	    if (n != -1)
		buf[n++] = d;
	    n_bits += nbits;
	}
    }
    void push_flag() {
	if (n > 0) {
	    pw.push(buf, n);
	}
	n_pkts++;
	n = 0;
    }

private:
    PacketWriter &pw;
    char buf[MTUMAX];
    int n;
public:
    int n_bits, n_errs, n_pkts;
};

enum modem_error {
    E_NONE,
    E_TOO_MANY_BAD_FRAMES,
    E_TOO_FEW_FRAMES,
    E_RX_ERROR,
    E_RX_WATCHDOG
};

// The Modem class.  This is a generic modem object, parametrizable
// with respect to system-dependent classes such as the interfaces
// to the audio and networking subsystems.

template <class PacketReader, class PWR, class AIO>
class Modem {
public:
    Modem(PacketReader &pr, PWR &pw, AIO &aio)
	: packet_reader(pr),
	  packet_writer(pw),
	  audio(aio),
	  txframer(packet_reader, tx_bitfifo),
	  transmitter(tx_bitfifo),
	  echo_canceller(audio, audio, transmitter, ec_rx_fifo),
	  packetizer(packet_writer),
	  deframer(packetizer),
	  receiver(ec_rx_fifo, rxbitfifo),
	  hssm(audio),
	  n_retrains(0),
          last_error(E_NONE)
    {
	audio.reset_gain();
	// Pre-load the transmitter buffer with silence to avoid
	// underflow while calculating the initial transmit signal block.
	for (int i = 0; i < 6000; i++)
	    audio.push(0.0f);
	audio.start_input();
	audio.start_output();
	init_done = true;
    }

    void reset() {
	hssm.reset();
	echo_canceller.reset();
	receiver.reset();
	transmitter.reset();
	audio.reset_gain();
	packetizer.reset();
	// XXX should reset audio somehow
	n_ticks = 0;
    }

    void retrain(modem_error err) {
        last_error = err;
#ifndef NO_RETRAIN
	reset();
	n_retrains++;
#endif
    }
    
    void tick() {
	//// Transmitter

	// Poll the packet source to see if transmit packets are available.
	packet_reader.poll();

	// Check for underrun
	(void) audio.playsize();
	
	// Calculate the number of transmit data bits needed to fill
	// the audio play buffers.
	int txbits = (int) (CHUNKSIZE * BITS_PER_SAMPLE) - tx_bitfifo.size();
	if (txbits < 0) txbits = 0;

	// Run the transmit framer long enough to generate those bits,
	// filling with HDLC flags if necessary.
	txframer.run(txbits);
	assert((int) tx_bitfifo.size() >= txbits);
	
	//// Receiver

	int audiosize = audio.size();
	if (audiosize > CHUNKSIZE)
	    audiosize = CHUNKSIZE;

	for (int i = 0; i < audiosize; i++) {
	    if (hssm.state != HS_STATE_END)
		audio.push(hssm.run(tofloat(audio())));
	    else {
		// XXX inefficient
		echo_canceller.wait = 5000 +
		    (hssm.identity == ID_A ? EC_SILENCE_TIME : 0);
		echo_canceller.run();
	    }
	}

	int watchdog = 0; // XXX
	while (ec_rx_fifo.size() >= 100) {
	    receiver.run();
	    if (++watchdog > 100000) {
		retrain(E_RX_WATCHDOG);
            	break;
            }
	}
	while (rxbitfifo.size()) {
	    deframer.run(rxbitfifo.pop(1));
	}
#ifndef AUTO_TESTING
	// Perform a simple sanity check on the HDLC framing output
	// and force a retrain if it fails.
	if (packetizer.n_bits + 8 * packetizer.n_pkts > 100000) {
	    // fprintf(stderr, "nerrs=%d, npkts=%d\n", packetizer.n_errs,
	    //  packetizer.n_pkts);
	    if (packetizer.n_errs >= 3)
	        retrain(E_TOO_MANY_BAD_FRAMES);
	    if (packetizer.n_pkts < 5) 
	        retrain(E_TOO_FEW_FRAMES);
	    packetizer.n_bits = 0;
	    packetizer.n_pkts = 0;
	    packetizer.n_errs = 0;
	}

	if (receiver.state == RX_ERROR) {
	    retrain(E_RX_ERROR);
	}
#endif
	n_ticks++;
    }

    PacketReader &packet_reader;
    PWR &packet_writer;

    AIO &audio;

    // Transmitter components

    typedef BitFIFO<8192> tx_bitfifo_t;
    
    tx_bitfifo_t tx_bitfifo;
    HDLCFramer<PacketReader, tx_bitfifo_t, typename PacketReader::PacketType>
	txframer;

    typedef Transmitter<sample_t, tx_bitfifo_t> transmitter_t;
	
    transmitter_t transmitter;

    // Echo canceller components 

    // A FIFO buffer is used between the echo canceller and the receiver.
    // XXX integrate with ec buffer
    typedef FIFO<sample_t, 4096> ec_rx_fifo_t;
    ec_rx_fifo_t ec_rx_fifo;
	
    EchoCanceller<sample_t, 16384, N_EC_TAPS, AIO, AIO,
	transmitter_t, ec_rx_fifo_t>
	echo_canceller;
    
    // Receiver components

    typedef BitFIFO<4096> rxbitfifo_t;
    rxbitfifo_t rxbitfifo;
    
    BitStreamPacketizer<PWR> packetizer;
    HDLCDeframer<BitStreamPacketizer<PWR> > deframer;

    Receiver<sample_t, ec_rx_fifo_t, rxbitfifo_t, N_EQ_TAPS>
	receiver;

    // Other components

    // The handshake state machine.  It is passed a reference
    // to the audio I/O handler object so that the HSSM can
    // tweak the ADC gain.
    HandshakeStateMachine hssm;
    
    int n_retrains;	     // The number of times we have forced a retrain
    modem_error last_error;  // Reason for latest retrain

    int n_ticks;

    bool init_done;
};

#endif
