#
# $Id: monitor.tcl,v 1.36 2006/10/18 13:20:35 gson Exp $
#
# GUI for real-time monitoring of modem parameters.
#
# (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
# Please refer to the file COPYRIGHT for detailed copyright information.
#

set daemon false

while {[llength $argv] && ![string compare [lindex $argv 0] "--daemon"]} {
    set daemon true
    set argv [lrange $argv 1 end]
}

# Decide between user and kernel mode and load the
# appropriate support code.

if {[llength $argv] > 0} {
    set mode kernel
    load ./audslmon_k.so
    set ifname [lindex $argv 0]
    set t KernelModeAudslModem
} else {
    set mode user
    load ./audslmon_u.so
    set t UserModeAudslModem
}

# Load support for trapping Unix signals and for running as a
# background daemon.

load ./signal.so
load ./daemonize.so

switch $mode {
    user {
	set shmidfile /var/run/audsld.shmid
	if {! [file exists $shmidfile]} {
	    puts stderr "no audsl daemon running, starting one..."
	    set child [exec ./audsld -s >@ stdout 2>@ stderr]
	}
	set f [open $shmidfile r]
	scan [gets $f] %x shmid
	close $f
	set modem [audsl_attach_shm $shmid]
    }
    kernel {
	if {[llength $argv] < 1} {
	    puts stderr "usage: monitor.tcl audslN"
	    exit 1
	}
	set modem [audsl_attach_kmem $ifname]
    }
}

set vars {
    hs_state hs_role rx_state
    ec_offset
    input_gain 
    post_ec_gain
    fqest
    rms_error
    tx_packets rx_packets
    tx_underruns rx_overruns
    n_retrains
    last_error
}

proc spectrum {modem v} {
    global t
    set ntaps [${t}_get_${v}_length $modem]
    # set pipe [open "| ./plot-spectrum.sh >/tmp/spectrum" w]
    set fn "${v}.taps"
    set pipe [open $fn w] 
    for {set i 0} {$i < $ntaps} {incr i} {
	puts $pipe \
	    "[${t}_get_${v}_tap $modem $i 0] [${t}_get_${v}_tap $modem $i 1]"
    }
    close $pipe
    exec sh plot-spectrum.sh --finite --complex $fn </dev/tty >/dev/tty &
}

proc update_gui_display {top modem} {
    global t
    # Only update the display if the modem has had time to initialize
    # itself; otherwise, we would end up following null pointers.
    if {[${t}_get_init_done $modem]} {

	# Update scalar variable displays

	global vars
	foreach v $vars {
	    $top.$v configure -text [${t}_get_$v $modem]
	}

	# Update canvas displays

	# The echo canceller is always real.
	set fdims(ec) {0}
	# The equalizer may be complex (in QAM mode).
	set fdims(eq) {1 0}

	# Display the real part and imaginary part in different colours.
	set color(0) blue
	set color(1) red

	foreach v {ec eq} {
	    set c $top.${v}taps 
	    set dims $fdims($v)
	    set ntaps [${t}_get_${v}_length $modem]
	    set tapmin -0.01
	    set tapmax 0.01
	    # Get the tap values and find the minimum/maximum
	    foreach dim $dims {
		set taps($dim) {}
		for {set i 0} {$i < $ntaps} {incr i} {
		    set tap [${t}_get_${v}_tap $modem $i $dim]
		    lappend taps($dim) $tap
		    if {$tap > $tapmax} {set tapmax $tap}
		    if {$tap < $tapmin} {set tapmin $tap}
		}
	    }
	    if {-$tapmin > $tapmax} {set tapmax [expr -$tapmin]}
	    set yscale [expr 50.0 / $tapmax]
	    $c delete curve
	    foreach dim $dims {
		set coords($dim) {}
		for {set i 0} {$i < $ntaps} {incr i} {
		    set x [expr 160 * $i / $ntaps]
		    set y [expr 50.0 - $yscale * [lindex $taps($dim) $i]]
		    lappend coords($dim) $x $y
		}
		eval $c create line $coords($dim) \
		    -tag curve -fill $color($dim)
	    }
	    $c itemconfigure range -text [format "%.02f" $tapmax]
	}
    }

    after 1000 [list update_gui_display $top $modem]
}

proc update_text_display {modem} {
    global t vars
    puts stdout ""
    # Only update the display if the modem has had time to initialize
    # itself; otherwise, we would end up following null pointers.
    if {[${t}_get_init_done $modem]} {
	foreach v $vars {
	    puts stdout [format "%-20s%s" $v [${t}_get_$v $modem]]
	}
    }
    after 1000 [list update_text_display $modem]    
}

if {$daemon} {
    daemon -nochdir
}

set exit_evloop 0

if {[llength [info commands wm]]} {
    # We are running under Tk; show the GUI.
    wm withdraw .
    set top [toplevel .audsl]
    wm title $top "audsl@[info hostname]"

    # option add *font 6x13
    option add *highlightThickness 0
    option add *padX 0
    option add *padY 0

    foreach v $vars {
	set f [frame $top.${v}frame -relief raised -borderwidth 2]
	pack [label $top.${v}legend -width 12 -text $v -anchor w] \
	    -in $f -side left 
	pack [label $top.$v -width 10 -anchor w] -in $f -side left
	pack $f
    }

    set canvastitle(ec) "Echo canceller"
    set canvastitle(eq) "Equalizer"

    foreach v {ec eq} {
	set c [canvas $top.${v}taps -relief raised -borderwidth 2 \
	       -highlightthickness 0 \
	       -width 160 -height 100]
	$c create line 0 50 160 50 -tag xaxis -fill black
	pack $c
	set button [button $c.b -text "save" -command [list spectrum $modem $v]]
	$c create window 160 100 -window $button -anchor se
	$c create text 4 4 -anchor nw -tag range
	$c create text 4 96  -anchor sw -text $canvastitle($v)
    }
    update_gui_display $top $modem
} else {
    # Tk not available; display in text mode.
    update_text_display $modem
    vwait exit_evloop
}
