// $Id: pam-rx.h,v 1.24 1999/03/04 15:38:53 gson Exp $
//
// A PR-PAM receiver.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#include "dsp.h"
#include "lfsr.h"
#include "pr.h"
#include "pam-protocol.h"
#include "resample-pull.h"
#include "scrambler.h"
#include "pll.h"
#include "sqrt.h"
#include "window-functions.h"
#include "pam-mapper.h"

//#define PLOT_RX_INPUT
//#define PLOT_RX_ERROR
//#define PLOT_COEFS

#if defined (PLOT_RX_INPUT) || defined(PLOT_RX_ERROR)
#include "plot-signal.h"
#endif

enum PRPAMReceiverState {
    PRPAM_RX_STATE_INITIAL,
    PRPAM_RX_STATE_HUNT,    // Hunting for fs/4 sync signal
    PRPAM_RX_STATE_SYNC,    // Tracking fs/4 sync signal
    PRPAM_RX_STATE_TRAIN,   // Training equalizer
    PRPAM_RX_STATE_DATA,    // Receiving data
    PRPAM_RX_STATE_ERROR    // Failed
};

#define NEXT(s) (state = s, nsamp = 0)

template<class T>
int sgn(T x) { return x >= 0 ? 1 : -1; }

inline float norm(float x) { return sqr(x); }
    
#define ADJUST_TIME 20

const float center_fq = 0.25 * 2 * M_PI; // Fs / 4

// This function object defines the impulse response of a lowpass
// FIR filter of length N, with a bandwidth of 97%.

template<int N>
class GaussianWindowedSinc97 {
public:
    double operator()(double x) {
	return sinc(x * 0.97) * GaussianWindow(3.0)(x / (N/2));
    }
};

// The PR-PAM receiver object.
// The input signal is obtained by calling a function object "inp" of
// class I without arguments; it should return a sample of type T.
// The received bits are delivered by calling output.push(),
// where "output" is an object of class O.
// The length of the adaptive equalizer is LMSLEN.

template<class T, class I, class O, int LMSLEN>
class PRPAMReceiver {
public:
    PRPAMReceiver(I &inp, O &outp)
	: input(inp), output(outp), mu(0.01),
	  pll(center_fq),
	  src(inp)
	  { reset(); }
    
    void reset() {
	state = PRPAM_RX_STATE_INITIAL;
	nsamp = 0;
	n_train = 0;
	pll.reset();
	fqest = center_fq;
	ampest = 0;
	mse = 0;
	equalizer_centered = false;
	gain = 1.0;
	equalizer.reset();
	training_sequence.reset();
	descrambler.reset();
	src.reset();
	fqadj = 0;
    }

    // Symbol rate (a.k.a. clock) synchronization.
    // This consists of two parts: synchronize1()
    // calculates a frequency adjustment value,
    // and need not be run for every sample.
    // synchronize2() applies the adjustment
    // and must be run for every sample.
    
    void synchronize1() {
	fqadj = (equalizer.w[LMSLEN/2-1] - equalizer.w[LMSLEN/2+1]) *
	    sgn(equalizer.w[LMSLEN/2]);
	if (! (fqadj > -1.0))
	    fqadj = -1.0;
	if (! (fqadj < 1.0))
	    fqadj = 1.0;
	if (isnan(fqadj) || isinf(fqadj)) {
	    fqadj = 0;
	    NEXT(PRPAM_RX_STATE_ERROR);	    
	}
    }

    void synchronize2() {
	src.phase += 0.000080 * fqadj;
    }
    
    // Run the receiver.  This nominally consumes a single input sample, but
    // in practice it may consume zero, one, or two samples due to sample 
    // rate conversion.
    
    void run();

    T get_input() {
	T isample = src();
	return isample * gain;
    }
    
    I &input;
    O &output;
    float mu;
    LMSFIR<T, T, LMSLEN, 100> equalizer;
    LFSR<BITS_PER_SAMPLE> training_sequence;
    Coder<T> coder;
    PRPAMReceiverState state;
    int nsamp;
    int n_train;
    PLL pll;
    float fqest;   // PLL frequency estimate, lowpass filtered
    float ampest;  // Input signal mean square amplitude estimate
    float mse;     // Mean square error estimate
    float save_fq;
    PAMMapper<BITS_PER_SAMPLE> mapper;
    PAMSlicer<BITS_PER_SAMPLE> slicer;
    PullSampleRateConverter<T, I, GaussianWindowedSinc97<32>, 32, 32, 5> src;
    V29Descrambler descrambler;
    bool equalizer_centered;
    float gain;
    float fqadj;
};

template<class T, class I, class O, int LMSLEN>
void PRPAMReceiver<T, I, O, LMSLEN>::run()
{
    T isample = get_input();
    
    switch (state) {

    case PRPAM_RX_STATE_INITIAL: {
	gain = 30;
	NEXT(PRPAM_RX_STATE_HUNT);
    }
	
    // Hunt for the synchronization carrier.

    case PRPAM_RX_STATE_HUNT: {
	ampest += 0.01f * ((isample * isample) - ampest);
	if (ampest > 0.5)
	    gain *= (1.0 - 0.01);

	(void) pll(isample);
	if (pll.locked())
	    NEXT(PRPAM_RX_STATE_SYNC);
	if (nsamp > 10 * 48000)
	    // Give up
	    NEXT(PRPAM_RX_STATE_ERROR);
	break;
   }

    // Track the synchronization carrier and try to estimate its frequency.
    // Also estimate the carrier amplitude, for setting the receiver gain.
    case PRPAM_RX_STATE_SYNC: {
	float osin = pll(isample);

	// Calculate a lowpass filtered frequency estimate
	fqest += 0.002f * (pll.osc.fq - fqest);
		
	// Calculate mean square amplitude estimate
	ampest += 0.01f * ((isample * isample) - ampest);
	    
	// Look for phase reversal
	if (osin * isample < pll.inphase * -0.4f) {
	    // Set the sample rate converter conversion factor
	    src.fq = center_fq / fqest;
	    // Set the receiver gain, taking into account the fact that 
	    // the peak amplitude of a sine wave is sqrt(2) times its RMS
	    // amplitude.
	    gain /= approx_sqrtf(ampest * 2.0f);
	    NEXT(PRPAM_RX_STATE_TRAIN);
	}

	break;
    }
	
    case PRPAM_RX_STATE_TRAIN: {
        // Wait until the equalizer has had time to achieve a resonable
	// convergence.  Then find the maximum tap magnitude, and carefully
	// nudge the sample rate converter phase to move the peak to the
	// center tap.

#ifdef PLOT_RX_INPUT
	static SignalPlot plot("rx_input");
	if (nsamp < 10000)
	    plot.push(isample);
#endif
	if (nsamp == 4000) {
	    // Find the maximum tap magnitude.
	    T max = -1;
	    int maxat = 0;
	    for (int j = 0; j < LMSLEN; j++) {
		float a = norm(equalizer.w[j]);
		if (a < 0)
		    a = -a;
		if (a > max) {
		    max = a;
		    maxat = j;
		}
	    }
	    // Estimate the true peak location by parabolic curve fitting
	    float s1 = norm(equalizer.w[maxat-1]);
	    float s2 = norm(equalizer.w[maxat]);
	    float s3 = norm(equalizer.w[maxat+1]);
	    float a = 0.5*(s3+s1-2*s2);
	    float b = 0.5*(s3-s1);
	    
	    float poserr = maxat + -b/(2*a) - LMSLEN/2;
	    if (! (poserr > -(LMSLEN/2) && poserr < (LMSLEN/2))) {
		poserr = 0;
		NEXT(PRPAM_RX_STATE_ERROR);
	    }
	    save_fq = src.fq;
	    // Adjust the sampling frequency to make the peak drift
	    // to the correct location in ADJUST_TIME sample times.
	    src.fq -= (src.fq * poserr / ADJUST_TIME);
	    //fprintf(stderr, "poserr = %f\n", poserr);
	} else if (nsamp == 4000 + ADJUST_TIME) {
	    // Done moving the peak.  Restore the sample rate converter speed,
	    // and reset the equalizer taps.
	    src.fq = save_fq;
	    equalizer.reset();
	}

	equalizer.push(isample);

	T d = equalizer.convolve();

	T comp = coder.convolve();
	T reconstruction = d - comp; // approximation of transmitter output

	// We effectively ignore the first LMSLEN/2 input samples,
	// so that the first training sequence bit will coincide with
	// the center sample in the equalizer when training begins.

	int training_bits;
	if (nsamp > LMSLEN/2) {
	    training_bits = training_sequence();
	    n_train++;
	} else {
	    training_bits = 0;
	}

	int bits = training_bits;

	T exact = mapper(bits);

	coder.push(exact);

	// error signal
	T e = reconstruction - exact;

#ifdef PLOT_RX_ERROR
	static SignalPlot errplot("rx_error");
	errplot.push(e);
#endif
	mse += 0.01 * (e*e - mse);

	if (n_train == N_TRAINING_SAMPLES-1) {
#ifdef AUTO_TESTING
	    NEXT(PRPAM_RX_STATE_ERROR);
#else
	    NEXT(PRPAM_RX_STATE_DATA);	    
#endif
	}

	// Adjust the equalizer taps on every sample so
	// that it converges quickly.
	equalizer.adjust(e * mu);

	// Wait for the equalizer to stabilize a bit, then
	// do symbol rate synchronization.
	if (nsamp > 5000) {
	    synchronize1();
	    synchronize2();
	}
	
#ifdef PLOT_COEFS
	static FILE *splot;
	if (! splot) {
	    char hostnamebuf[100];
	    gethostname(hostnamebuf, sizeof(hostnamebuf));
	    splot = fopen(hostnamebuf, "w");
	}
	if (nsamp % 100 == 0 && nsamp < 10000) {
	    for (int i = 0; i < equalizer.length(); i++) {
		fprintf(splot, "%f\n", equalizer.w[i]);
	    }
	    fprintf(splot, "\n");
	    fflush(splot);
	}
#endif

	break;
    }

    case PRPAM_RX_STATE_DATA: {
	equalizer.push(isample);

	T d = equalizer.convolve();

	T comp = coder.convolve();
	T reconstruction = d - comp; // approximation of transmitter output

	int bits = slicer(reconstruction);
	T exact = mapper(bits);

	coder.push(exact);

	// error signal
	T e = reconstruction - exact;

#ifdef PLOT_RX_ERROR
	static SignalPlot errplot("rx_error");
	errplot.push(e);
#endif

	for (int j = 0; j < BITS_PER_SAMPLE; j++) {
	    int bit = (bits >> j) & 1;
	    output.push(descrambler(bit), 1);
	}

	// Update the error estimate, equalizer taps and frequency 
	// adjustment value only occasionally now that the initial 
	// training is finished.
	if ((nsamp & 0x0F) == 0) {
	    mse += 0.01 * (e*e - mse);
	    equalizer.adjust(e * mu);
	    synchronize1();
	}
	synchronize2();

	break;
    }

    case PRPAM_RX_STATE_ERROR: {
	break;
    }

    }
    nsamp++;
};
