// $Id: pll.h,v 1.16 2006/10/17 09:42:36 gson Exp $
//
// Phase-locked loop class.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#ifndef _PLL_H
#define _PLL_H

#include <math.h>

#include "sincos.h"

// An oscillator with sine and cosine outputs.  The frequency
// is given in radians per sample.

class Osc {
public:
    Osc(double f) : fq(f) { reset(); }

    void reset() { w = 0; }

    float sin() { return sin2pi(w * (1.0/(2*M_PI))); }
    float cos() { return cos2pi(w * (1.0/(2*M_PI))); }
    void step() {
	w += fq;
	if (w > (float) (2*M_PI))
	    w -= (float) (2*M_PI);
    }
    float operator()() { float r = sin(); step(); return r; }
    double fq; // Frequency
    double w;  // Current phase angle
};

class PLL {
public:
    PLL(double cfq) : center_fq(cfq), osc(center_fq) { reset(); }

    void reset() {
	osc.reset();
	osc.fq = center_fq;
	inphase = outphase = 0.0;
	n_sync = 0;
    }
    
    float operator()(float d) {
	float ocos = osc.cos();
	float osin = osc.sin();

	// PLL adjustment
	osc.fq += 0.01f * (ocos * d) - 0.01f * (osc.fq - center_fq);
	inphase += 0.01f * (osin * d - inphase);
	outphase += 0.01f * (ocos * d - outphase);
	if (inphase > 4.0f * fabs(outphase)) {
	    if (! locked()) {
		n_sync++;
	    }
	} else {
	    n_sync = 0;
	}

	osc.step();
	return osin;
    }
    
    bool locked() { return n_sync == 1000; }

    double center_fq;
    Osc osc;
    float inphase; // in-phase component correlation
    float outphase; // out-of-phase component correlation
    int n_sync;
};

#endif
