// $Id: polyphase-fir.h,v 1.4 1999/06/29 18:27:43 gson Exp $
//
// Polyphase filter for rational resampling. 
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

// A polyphase FIR filter.  Decimates the input signal by a factor
// of freq/NPHASES, or interpolates by a factor of NPHASES/freq,
// where freq and NPHASES are integers.  

#ifndef _POLYPHASE_FIR_H
#define _POLYPHASE_FIR_H

#include "dsp.h"

// "Kernel" is a continuous function defining
// the impulse response of the filter at the input samping rate.
//
// When interpolating, it will typically be a lowpass filter with
// a cutoff frequency of fs/2, e.g., a windowed sinc(x).
// When decimating, the cutoff frequency should be correspondingly
// lower to avoid aliasing.

template <class T, class Kernel, int NTAPS, int NPHASES>
class PolyphaseFIR: public ContiguousSignalBuffer<T, NTAPS, 256> {
public:
    typedef ContiguousSignalBuffer<T, NTAPS, 256> Base;

    PolyphaseFIR(const Kernel &k, int freq)
        : fq(freq)
    {
	init_taps(k); reset();
    }

    void reset() {
	phase = 0;
    }

    T filter() {
	return ::convolve(*this, taps[phase], T(0.0f), NTAPS); 
    }

    void init_taps(const Kernel &k) {
	for (int j = 0; j < NPHASES; j++) {
	    double phase = (double) j / NPHASES;
	    for (int i = 0; i < NTAPS; i++) {
		double s = i - (NTAPS/2) + phase;
		taps[j][i] = k(s) * 0.6; // Gibb
	    }
	}
    }

    int fq;
    int phase;

    float taps[NPHASES][NTAPS];
};

template <class T, class I, class Kernel, int NTAPS,
    int NPHASES>
class PullPolyphaseFIR:
public PolyphaseFIR<T, Kernel, NTAPS, NPHASES>
{
public:
    PullPolyphaseFIR(I &input, Kernel k, int fq)
	: PolyphaseFIR<T, Kernel, NTAPS, NPHASES>(k, fq), inp(input) { }

    T operator()() {
	while (phase >= NPHASES) {
	    Base::push(inp());
	    phase -= NPHASES;
	}
	T r = filter();
	phase += fq;
	return r;
    }

private:
    I &inp;
};

#endif
