// $Id: pulse-shaper.h,v 1.15 2004/11/26 18:59:12 gson Exp $
//
// Pulse shapers for full response QAM and PAM transmitters.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

// The QAM pulse shaper is hardcoded for a symbol rate of one
// two-dimensional symbol per three signal samples, and the PAM
// pulse shaper is hardcoded for two one-dimensional symbol
// samples per three signal samples.

#ifndef _PULSE_SHAPER_H
#define _PULSE_SHAPER_H

#include <complex>

using std::complex;

#include "dsp.h"
#include "window-functions.h"
#include "parameters.h"

//const int sidebits = 2;
//const int nbits = 1 + sidebits * 2;

#define sidebits PULSEWIDTH
#define nbits (1 + sidebits * 2)

// This function object defines the transmit pulse shape.
// Currently, we use the classic root raised cosine shape.

class SqrtRCPulseShape {
public:
    SqrtRCPulseShape(double b)
	: beta(b) { }
    inline double operator()(double x) {
	//return sinc(x) * GaussianWindow(PULSEROLLOFF)(x * 1.0 / (sidebits+1));
	//return sinc(x) * RaisedCosineWindow()(x * 1.0 / (sidebits+1));
	//return sinc(x) * ProakisWindow(PULSEROLLOFF)(x);
	x += 0.00001; // Avoid division by zero
	double t = 
	    (cos ((1 + beta) * M_PI * x) +
	     (1.0 / (4 * beta)) * sin((1 - beta) * M_PI * x) / x) /
	     (1 - sqr(4 * beta * x));
	return (4 * beta / M_PI) * t;
    }
    double beta;
};

// Specialize the above for a beta value of 0.3.

class SqrtRCPulseShape03: public SqrtRCPulseShape {
public:
    SqrtRCPulseShape03()
	: SqrtRCPulseShape(0.3) { }
};

// A table of coefficients for a table driven pulse shaper.
// The pulse amplitude is neg_amp for "0" bits and pos_amp
// for "1" bits. 

template <class PulseShape>
class PulseShaperCoefs {
public:
    PulseShaperCoefs(float neg_amp, float pos_amp, float width)  {
	PulseShape pulse;
	// Loop through the three possible phases
	for (int ph = 0; ph < 3; ph++) {
	    // t is the phase offset for this phase: -1/3, 0, or +1/3
	    float t = (float) (ph - 1) / 3;
	    // Loop through all possible bit combinations for the
	    // output bits that affect the current sample
	    for (int i = 0; i < (1 << nbits); i++) {
		float sum = 0;
		for (int n = 0; n < nbits; n++) {
		    float p = pulse(width * (t - (n - sidebits)));
		    if ((i & (1 << n)) == 0)
			p *= neg_amp;
		    else
			p *= pos_amp;
		    sum += p;
		}
		table[ph][i] = sum;
	    }
	}
    }
    float table[3][1 << nbits];
};


// A QAMPulseShaper generates complex signal samples from input data
// bits.  The current sample value is read as a complex number through
// operator(), or as separate real and imaginary components through
// real() and imag().  To step to the next sample, call step().
// The input data is obtained by occasionally calling the input object
// I; its operator() should return two bits of data at a time.

template <class I>
class QAMPulseShaper {
public:
    QAMPulseShaper(I &input, float neg_amp, float pos_amp)
	: inp(input), coefs(neg_amp, pos_amp, 1.0) { reset(); }

    void reset() {
	phase = 0;
	re_bits = 0;
	im_bits = 0;
    }

    void step() {
	if (++phase >= 3) {
	    int bits = inp();
	    re_bits |= ((bits >> 0) & 1) << nbits;
	    re_bits >>= 1;
	    im_bits |= ((bits >> 1) & 1) << nbits;
	    im_bits >>= 1;
	    phase = 0;
	}
    }

    float real() { return coefs.table[phase][re_bits]; }
    float imag() { return coefs.table[phase][im_bits]; }
    
    complex<float> operator()() {
	complex<float> r(real(), imag());
	step();
	return r;
    }

    I &inp;
    PulseShaperCoefs<SqrtRCPulseShape03> coefs; // XXX static
    int phase;
    int re_bits;
    int im_bits;
};

// Full-response PAM pulse shaper.

template <class I>
class PAMPulseShaper {
public:
    PAMPulseShaper(I &input, float neg_amp, float pos_amp)
	: inp(input), coefs(neg_amp, pos_amp, 1.0) { reset(); } 

    void reset() {
	phase = 0;
	bits = 0;
    }

    void step() {
	phase += 2;
	if (phase >= 3) {
	    int in = inp();
	    bits |= (in & 0x01) << nbits;
	    bits >>= 1;
	    phase -= 3;
	}
    }
    float operator()() {
	float t = coefs.table[phase][bits];
	step();
	return t;
    }
    I &inp;
    PulseShaperCoefs<SqrtRCPulseShape03> coefs;
    int phase;
    int bits;
};

// Staggered QAM pulse shaper.  The SQAM signal basically consists
// of two interleaved PAM signals, each modulated by an fs/4 sine
// wave.   

// XXX share coefs between s0, s1

template <class I>
class SQAMPulseShaper {
public:
    SQAMPulseShaper(I &inp, float neg_amp, float pos_amp)
	: s0(inp, neg_amp, pos_amp), s1(inp, neg_amp, pos_amp) { reset(); }
    void reset() { carrier_phase = 0; }
    float operator()() {
	float r;
	switch (carrier_phase) {
	case 0: r = s0(); break;
	case 1: r = s1(); break;
	case 2: r = -s0(); break;
	case 3: r = -s1(); break;
	}
	carrier_phase++;
	carrier_phase &= 0x03;
	return r;
    }
    PAMPulseShaper<I> s0;
    PAMPulseShaper<I> s1;
    int carrier_phase;
};

#undef sidebits
#undef nbits

#endif
