// $Id: qam-rx.h,v 1.37 2004/11/26 18:59:12 gson Exp $
//
// QAM receiver.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#include <stdio.h>
#include <complex>

using std::complex;

#include "qam-protocol.h"
#include "qam-mapper.h"
#define COMPLEX_SRC // XXX
#include "resample-push.h"
#include "window-functions.h"
#include "pll.h"
#include "sqrt.h"
#include "hilbert.h"
#include "parameters.h"
#include "pulse-shaper.h"

//#define PLOT_RX_INPUT		// Plot receiver input signal
//#define PLOT_RX_ERROR		// Plot receiver error residual
//#define PLOT_COEFS		// Plot equalizer coefficients

#if defined(PLOT_COEFS) || defined(PLOT_RX_ERROR) || defined(PLOT_RX_INPUT)
#include "plot-signal.h"
#endif
#ifdef PLOT_COEFS
#include <stdio.h> // XXX
#endif

enum ReceiverState {
    QAM_RX_STATE_INIT,    // Not running yet
    QAM_RX_STATE_HUNT,    // Hunting for fs/4 sync signal
    QAM_RX_STATE_SYNC,    // Tracking fs/4 sync signal
    QAM_RX_STATE_TRAIN,   // Training equalizer
    QAM_RX_STATE_DATA,    // Receiving data
    QAM_RX_STATE_ERROR    // Failed
};

template <class T, class I, class O, int LMSLEN> 
class QAMReceiver;

// The baseband part of the receiver.  Its push() method
// is called with a complex baseband sample (of type complex<T>)
// once every symbol period.

#define NEXT(s) ( rx.state = s, nsamp = -1 )

template<class T>
int sgn(T x) { return x >= 0 ? 1 : -1; }

#define ADJUST_TIME 20

template <class T, class I, class O, int LMSLEN>
class BasebandPart {
public:
    BasebandPart(QAMReceiver<T, I, O, LMSLEN> &receiver)
	: rx(receiver) { reset(); }

    void reset() {
	training_sequence.reset();
	mapper.reset();
	nsamp = 0;
	n_train = 0;
	descrambler.reset();
	filter.reset();
	mse = 0;
	adjf = cadjf = 0;
    }

    void synchronize() {
	// Symbol rate (a.k.a. clock) synchronization
	float poserr =
	    (filter.w[LMSLEN/2-1].real() - filter.w[LMSLEN/2+1].real()) *
	    sgn(filter.w[LMSLEN/2].real()) +
	    (filter.w[LMSLEN/2-1].imag() - filter.w[LMSLEN/2+1].imag()) *
	    sgn(filter.w[LMSLEN/2].imag());
	rx.src.phase += 0.000800 * poserr;
	rx.src.fq += 0.000000200 * poserr;
	
	// Carrier synchronization
	rx.pll.osc.w -= 0.000800 * filter.w[LMSLEN/2].imag();
	rx.pll.osc.fq -= 0.000000200 * filter.w[LMSLEN/2].imag();
    }
    
    void push(complex<T> input) {

	switch (rx.state) {
	    
	case QAM_RX_STATE_TRAIN: {

	    // Wait until the equalizer has had time to achieve a resonable
	    // convergence.  Then find the maximum tap magnitude, and carefully
	    // nudge the sample rate converter phase to move the peak to the
	    // center tap.
	    
	    if (nsamp == 4000) {
		
		// Find the maximum tap magnitude.
		T max = -1;
		int maxat = 0;
		for (int j = 0; j < LMSLEN; j++) {
		    float a = norm(filter.w[j]);
		    if (a < 0)
			a = -a;
		    if (a > max) {
			max = a;
			maxat = j;
		    }
		}
		// Estimate the true peak location by parabolic curve fitting
		float s1 = norm(filter.w[maxat-1]);
		float s2 = norm(filter.w[maxat]);
		float s3 = norm(filter.w[maxat+1]);
		float a = 0.5*(s3+s1-2*s2);
		float b = 0.5*(s3-s1);
		float poserr = maxat + -b/(2*a) - LMSLEN/2;
		save_fq = rx.src.fq;
		// Adjust the sampling frequency to make the peak drift
		// to the correct location in ADJUST_TIME sample times.
		rx.src.fq -= (rx.src.fq * poserr / ADJUST_TIME);
		//fprintf(stderr, "poserr = %f\n", poserr);
	    } else if (nsamp == 4000 + ADJUST_TIME) {
		// Done moving the peak.  Restore the sample rate converter
		// speed, and reset the equalizer taps.
		rx.src.fq = save_fq;
		filter.reset();
	    }

	    filter.push(input);
	
	    complex<T> d = filter.convolve();
	    complex<T> exact;

	    // We effectively ignore the first LMSLEN/2 input samples,
	    // so that the first training sequence bit will coincide with
	    // the center sample in the equalizer when training begins,
	    // at least approximately.  The clock synchronizer takes
	    // care of any remaining error as long as the correlation 
	    // peak is well within the equalizer.
	    
	    if (nsamp > (LMSLEN/2) +
		(rx.src.delay() + rx.hilbert.delay()) * 16 / 48 + 1)
	    {
		exact = mapper(training_sequence());
		++n_train;
	    } else {
		exact = 0.0;
	    }
	    if (n_train == N_TRAINING_SYMBOLS)
		NEXT(QAM_RX_STATE_DATA);

	    complex<T> error = d - exact;

	    mse += 0.001 * (norm(error) - mse);

	    filter.adjust(error * 0.1f);

	    if (nsamp > 5000)
		synchronize();

#ifdef PLOT_RX_ERROR	
	    static SignalPlot plot("rxerr");
	    plot.push(error.real()*0.2);
	    plot.push(error.imag()*0.2);
#endif

#ifdef PLOT_COEFS
	    if (nsamp % 500 == 0) {
		static FILE *splot;
		if (! splot) splot = open_log_file("eq.taps");
		for (int i = 0; i < filter.length(); i++) {
		    fprintf(splot, "%f %f %f\n",
			    sqrt(norm(filter.w[i])),
			    filter.w[i].real(), filter.w[i].imag());
		}
		fprintf(splot, "\n");
		fflush(splot);
	    }
#endif
	    break;
	}
	    
	case QAM_RX_STATE_DATA: {
	    filter.push(input);
#if 1
	    complex<T> d = filter.convolve();
#else
	    // Run the filter, using only the real coefficients.
	    // The complex ones should have adapted to zero by now.
	    complex<T> sum = 0;
	    for (int j = 0; j < filter.length(); j++)
		sum += filter[j] * filter.w[j].real();

	    complex<T> d = sum;
#endif
	    int bits = slicer(d);
	    for (int j = 0; j < N_BITS_PER_SYMBOL; j++) {
		int bit = (bits >> j) & 1;
		rx.out.push(descrambler(bit), 1);
	    }

	    // Adjust eq taps and clock / carrier frequencies only
	    // occasionally.
	    if ((nsamp & 0x0F) == 0) {
		// Equalize in decision directed mode.
		complex<T> exact = mapper(bits);
		// We must do a full complex convolution when adapting,
		// even if we otherwise use only the real coefficients.
		//complex<T> error = filter.convolve() - exact;
		//-not
		complex<T> error = d - exact;
		mse += 0.016 * (norm(error) - mse);
		filter.adjust(error * 0.1f);
		synchronize();
	    }

	    break;
	}
	    
	}
	++nsamp;
    }

    int nsamp;
    int n_train;
    QAMTrainingSequence training_sequence;
    QAMSlicer<T, N_BITS_PER_SYMBOL> slicer;
    QAMMapper<T, N_BITS_PER_SYMBOL> mapper;
    LMSFIR<complex<T>, complex<T>, LMSLEN, 100> filter;
    QAMReceiver<T, I, O, LMSLEN> &rx;
    QAMDescrambler descrambler;
    float mse;
    float save_fq;
    float adjf, cadjf;
};

#undef NEXT

#define NEXT(s) ( state = s, state = s, nsamp = -1)

const float center_fq = 0.25 * 2 * M_PI; // Fs / 4

class MatchedFilterResponse {
public:
    SqrtRCPulseShape03 pulse;
    double operator()(double x) {
	return (1.0 / 3.0) * pulse(x / 3.0);
    }
};

// The passband part of the receiver.

template <class T, class I, class O, int LMSLEN> 
class QAMReceiver {
public:
    QAMReceiver(I &input, O &output)
	: in(input), out(output), pll(center_fq), baseband_part(*this),
	  src(baseband_part)
    {
	reset();
	// Sample the 48 kHz signal at the 16 kHz symbol rate
	src.fq = 48.0 / 16.0;
	gain = 1.0;
    }
    
    void reset() {
	state = QAM_RX_STATE_INIT;
	nsamp = 0;
	pll.reset();
	fqest = center_fq;
	ampest = 0;
	src.reset();
	baseband_part.reset();
    }
    
    void run() {
	T input = (T) in() * gain;

#ifdef PLOT_RX_INPUT
	static SignalPlot plot("rxin");
	plot.push(input * 0.5);
#endif

	switch (state) {
	case QAM_RX_STATE_INIT: {
	    NEXT(QAM_RX_STATE_HUNT);
	    break;
	}

	case QAM_RX_STATE_HUNT: {
	    (void) pll(input);
	    if (pll.locked())
		NEXT(QAM_RX_STATE_SYNC);
	    if (nsamp > 10 * 48000)
		NEXT(QAM_RX_STATE_ERROR);
	    break;
	}

	case QAM_RX_STATE_SYNC: {
	    float d = input;
	    float osin = pll(d);
	    
	    // Calculate a lowpass filtered frequency estimate
	    fqest += 0.002f * (pll.osc.fq - fqest);
	    
	    // Calculate mean square amplitude estimate
	    ampest += 0.01f * ((d * d) - ampest);
	    
	    // Look for phase reversal
	    if (osin * d < pll.inphase * -0.2f) {
		// fprintf(stderr, "reversal at %d; fq estimate = %g\n",
		//      nsamp, center_fq / fqest);
		// fprintf(stderr, "src.fq = %g, new = %g\n", src.fq,
		//      (center_fq/fqest)*(48.0/16.0));
		src.fq *= (center_fq / fqest);
		
		// Set the receiver gain, taking into account the fact that 
		// the peak amplitude of a sine wave is sqrt(2) times its RMS
		// amplitude.
		gain /= approx_sqrtf(ampest * 2.0f);

		// Use our best estimate for the initial oscillator frequency.
		// pll.osc.fq = fqest;
		pll.osc.fq = center_fq;

		NEXT(QAM_RX_STATE_TRAIN);
	    }
	    break;
	}

	case QAM_RX_STATE_TRAIN:
	case QAM_RX_STATE_DATA: {
	    complex<T> analytic(hilbert.split(input));
	    complex<T> baseband(analytic *
			conj(complex<T>(pll.osc.cos(), pll.osc.sin())));
	    src.push(baseband);
	    pll.osc.step();
	    break;
	}

	case QAM_RX_STATE_ERROR: {
	    break;
	}
	    
	}
	++nsamp;
    }

    I &in;
    O &out;
    int state;
    int nsamp;
    PLL pll;
    float fqest;   // PLL frequency estimate, lowpass filtered
    float ampest;  // Input signal mean square amplitude estimate
    HilbertTransformer<T, 31> hilbert;
    Delay<T, 15> delay;    
    BasebandPart<T, I, O, LMSLEN> baseband_part;
    PushSampleRateConverter<complex<T>, BasebandPart<T, I, O, LMSLEN>,
        MatchedFilterResponse, SRCTAPS, 64, 6> src;
    float gain;
};

#undef NEXT
