// $Id: qam-tx.h,v 1.15 1999/05/30 19:38:25 gson Exp $
//
// QAM transmitter.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#include <complex>

#include "qam-protocol.h"
#include "pulse-shaper.h"
#include "window-functions.h"
#include "pll.h"
#include "qam-mapper.h"

enum TransmitterState {
    TX_STATE_TRAIN_EC,
    TX_STATE_PILOT,
    TX_STATE_REVERSAL,
    TX_STATE_TRAIN,
    TX_STATE_DATA
};

#define NEXT(s) (state = s, nsamp = -1)

// Calculate the data bits to use in the next baseband signal sample
// (by calling operator()).
// The object "input" (of class I) is the source of bits to transmit
// (after the pilot and training sequences are finished).

template <class I>
class TransmitSequencer {
public:
    TransmitSequencer(I &input)
	: inp(input) { reset(); }

    void reset() {
	state = TX_STATE_PILOT;
	nsamp = 0;
	scrambler.reset();
    }

    int operator()() {
	int bits;
	switch (state) {
	case TX_STATE_TRAIN_EC: {
	    bits = training_sequence();
	    break;
	}
	case TX_STATE_PILOT: {
	    bits = 0x00;
	    if (nsamp == N_PILOT_SYMBOLS-1)
		NEXT(TX_STATE_REVERSAL);
	    break;
	}
	case TX_STATE_REVERSAL: {
	    bits = 0x03;
	    if (nsamp == N_REVERSAL_SYMBOLS-1)
		NEXT(TX_STATE_TRAIN);
	    break;
	}
	case TX_STATE_TRAIN: {
	    bits = training_sequence();	    
	    if (nsamp == N_TRAINING_SYMBOLS-1)
		NEXT(TX_STATE_DATA);
	    break;
	}
	default: // avoid compiler warning
	case TX_STATE_DATA: {
	    bits = 0;
	    for (int j = 0; j < N_BITS_PER_SYMBOL; j++) {
		bits |= scrambler(inp.pop()) << j; // XXX bit at a time
	    }
	    break;
	}
	}
	++nsamp;
	return bits;
    }
    
    I &inp;
    TransmitterState state;
    int nsamp;

    QAMTrainingSequence training_sequence;
    QAMScrambler scrambler;
};


template <class T, class I>
class QAMTransmitter {
public:
    QAMTransmitter(I &input)
	: seq(input), shaper(seq, -0.4, 0.4)
    {
	// At fs = 48 kHz, the symbol rate is 16 kHz.
	//src.fq = 16.0/48.0;
	reset();
    }
    void reset() {
	seq.reset();
	shaper.reset();
	phase = 0;
    }
    void training_mode(bool on) {
	if (on)
	    seq.state = TX_STATE_TRAIN_EC;
	else
	    seq.state = TX_STATE_PILOT;
	seq.training_sequence.reset();
    }
    T operator()() {
	T r; 
	switch (phase) {
	case 0: r = shaper.real(); break;
	case 1: r = -shaper.imag(); break;
	case 2: r = -shaper.real(); break;
	case 3: r = shaper.imag(); break;
	default: r = 0; assert(0); break;
	}
	shaper.step();
	phase = (phase+1) & 0x03; // Phase wraps modulo 4
	return r;
    }

    TransmitSequencer<I> seq;
    QAMPulseShaper<TransmitSequencer<I> > shaper;
    int phase;
};
