/* $Id: signal.c,v 1.4 2004/11/26 18:08:22 gson Exp $ */

/*
 * Simple Tcl extension for sending and handling Unix signals.
 *
 * Examples:
 *
 *   kill 1 42
 *
 * This will send signal 1 (hangup) to process 42.
 *
 *   signal 1 "puts stdout HUP"
 *
 * This will cause the text HUP to be printed whenever this
 * process gets a hangup signal, instead of exiting.
 *
 */

#include <signal.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "tcl.h"

static Tcl_Interp *sig_interp;
static char *handlers[NSIG];

static void sighandler(int signo)
{
    Tcl_Eval(sig_interp, handlers[signo]);
}

static int
signal_proc(ClientData client_data, Tcl_Interp *ip, int argc, const char **argv)
{
    int r;
    int signo;
    if (argc != 3) {
	Tcl_AppendResult(ip, "signal: wrong # args", (char *) 0);
	return TCL_ERROR;
    }
    r = Tcl_GetInt(ip, argv[1], &signo);
    if (r != TCL_OK) return r;
    if (signo < 0 || signo >= NSIG) {
	Tcl_AppendResult(ip, "signal ", argv[1], " out of range", (char *) 0);
	return TCL_ERROR;
    }
    if (handlers[signo])
	free(handlers[signo]);
    handlers[signo] = malloc(strlen(argv[2]) + 1);
    strcpy(handlers[signo], argv[2]);
    sig_interp = ip;
    signal(signo, sighandler);
    return TCL_OK;
}

static int
kill_proc(ClientData client_data, Tcl_Interp *ip, int argc, const char **argv)
{
    int r;
    int signo, pidno;
    if (argc != 3) {
	Tcl_AppendResult(ip, "kill: wrong # args", (char *) 0);
	return TCL_ERROR;
    }
    r = Tcl_GetInt(ip, argv[1], &signo);
    if (r != TCL_OK) return r;
    r = Tcl_GetInt(ip, argv[2], &pidno);
    if (r != TCL_OK) return r;
    r = kill(pidno, signo);
    if (r < 0) {
	Tcl_AppendResult(ip, "kill: ", strerror(errno), (char *) 0);
    }
    return TCL_OK;
}

int Signal_Init (Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "signal", signal_proc, 0, 0);
    Tcl_CreateCommand(interp, "kill", kill_proc, 0, 0);
    return TCL_OK;
}

