#ifndef _FREEBSD_AUDIO_H
#define _FREEBSD_AUDIO_H

#include <sys/types.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>

#include "dsp.h"
#include "gain.h"

#define BS 65536

#define MASTER_LEVEL 90
#define RECORD_LEVEL 90
#define PCM_LEVEL 90

class FreeBSDAudioIO: public Gain {
public:
    // Initialize the audio device, configure it 
    // for full duplex operation, set an appropriate
    // (small) block size, etc.
    
    FreeBSDAudioIO(int audio_fd) 
	: fd(audio_fd),
	  input_gain(0),
	  output_gain(0.35)
    {
	roff = poff = 0;
	rdmaoff = pdmaoff = 0;
	rbuffersize = pbuffersize = BS;

	snd_capabilities capmask;
	if (ioctl(fd, AIOGCAP, &capmask) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_GETFMTS: %s\n", strerror(errno));
	    exit(1);
	}
	if (!(capmask.formats & AFMT_FULLDUPLEX)) {
	    fprintf(stderr, "Full duplex mode not supported.\n");
	    exit(1);
	}

	// Set the block size (a.k.a. fragment size)
	// 7FFF = no limit on number of blocks
	// 0008 = block size 2^8 
	int fragspec = 0x7FFF0008;
	if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &fragspec) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_SETFRAGMENT: %s\n", strerror(errno));
	    exit(1);
	}

	int format = AFMT_S16_LE;  
	if (ioctl(fd, SNDCTL_DSP_SETFMT, &format) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_SETFMT: %s\n", strerror(errno));
	    exit(1);
	}
	if (format != AFMT_S16_LE) {
	    fprintf(stderr, "audio device does not support the "
		    "S16_LE audio format\n");
	    exit(1);
	}
	    
	int stereo = 0;
	if (ioctl(fd, SNDCTL_DSP_STEREO, &stereo) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_STEREO: %s\n", strerror(errno));
	    exit(1);
	}
	if (stereo != 0) {
	    fprintf(stderr, "Audio device does not support mono mode\n");
	    exit(1);	    
	}

	int wanted_speed = 48000;
	int speed = wanted_speed;
	if (ioctl(fd, SNDCTL_DSP_SPEED, &speed) == -1)  {
	    fprintf(stderr, "SNDCTL_DSP_SPEED: %s\n", strerror(errno));
	    exit(1);
	}
	if (speed > wanted_speed * 11 / 10 || speed < wanted_speed * 9 / 10) {
	    fprintf(stderr, "Audio device does not support a "  
		    "sampling rate of %d Hz (closest %d Hz)\n",
		    wanted_speed, speed);
	    exit(1);
	}

	// Set suitable gains for the signal sources we actually use
	mixer_ioctl(SOUND_MIXER_WRITE_PCM, "DAC gain", 
		    PCM_LEVEL);
	mixer_ioctl(SOUND_MIXER_WRITE_VOLUME, "master volume", 
		    MASTER_LEVEL);
	mixer_ioctl(SOUND_MIXER_WRITE_RECLEV, "Record level", 
		    RECORD_LEVEL);

	// Turn everything else off to avoid picking up unnecesary noise
	mixer_ioctl(SOUND_MIXER_WRITE_CD, "CD gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_LINE1, "LINE1 gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_MIC, "MIC gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_LINE, "LINE gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_SPEAKER, "Speaker gain", 0);

	// Record from LINE IN only
	mixer_ioctl(SOUND_MIXER_WRITE_RECSRC, "record source", 1 << SOUND_MIXER_LINE);
	
	int recsources = 1 << SOUND_MIXER_LINE;
	if (ioctl(fd, SOUND_MIXER_WRITE_RECSRC, &recsources) == -1) {
	    fprintf(stderr, "SOUND_MIXER_WRITE_RECSRC: %s\n", strerror(errno));
	    exit(1);
	}
	// Set non-blocking mode 
	int r = fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) | O_NONBLOCK);
	if (r < 0) {
	    fprintf(stderr, "fcntl O_NONBLOCK: %s\n", strerror(errno));
	    exit(1);
	}
      
	while (read(fd, rbuf, rbuffersize) >= 1)
	    ;
    }

    // Start recording, if your system does not
    // start it automatically when the audio device
    // is opened.
    
    void start_input() {
	flip_trigger(PCM_ENABLE_INPUT);
    }

    // Start playback, if your system does not
    // start it automatically when the audio device
    // is opened.
    
    void start_output() {
	flip_trigger(PCM_ENABLE_OUTPUT);      
    }

    int _playsize() {
	int sizebytes = poff - pdmaoff;
	if (sizebytes < 0)
	    sizebytes += pbuffersize;
	return sizebytes;
    }

    // Get the number of unplayed samples in the play queue.
    int playsize() {
	int sizebytes = _playsize();
	if (sizebytes > (pbuffersize * 3 / 4))
	    underruns++;
    
	int sizesamples = sizebytes / 2;
	return sizesamples;
    }


    void push(float f) {
	assert(_playsize() < pbuffersize - 4);

	s16 s = to16(f * output_gain);  
	pbuf[poff+0] = s.v & 0xFF;
	pbuf[poff+1] = s.v >> 8;
	poff += 2;
	if (poff == pbuffersize)
	    poff = 0;

	// If we have a good-sized chunk of unplayed data, (try to) play it.
	int toplay = _playsize();

	if (toplay >= 256) {
	    while (toplay > 0) {
		// Number of bytes to play this iteration
		int playnow = toplay; 
		// Number of bytes before buffer wraps
		int wrap_at = pbuffersize - pdmaoff; 
		if (playnow > wrap_at)
		    playnow = wrap_at;
		int nplay = write(fd, pbuf + pdmaoff, playnow);
		if (nplay < 0 && errno == EWOULDBLOCK)
		    nplay = 0;
		assert (nplay >= 0);
		pdmaoff += nplay;
		if (pdmaoff == pbuffersize)
		    pdmaoff = 0;
		toplay -= nplay;
		// No use trying to play more if this play was short.
		if (nplay < playnow)
		    break;
	    }
	}
    }

    int _size() {
	int sizebytes = rdmaoff - roff;
	if (sizebytes < 0)
	    sizebytes += rbuffersize;
	return sizebytes;
    }

    int size() {
	// First try to fill the buffer by doing nonblocking
	// reads from our file descriptor.
	// Try to make the buffer half full.
	int toread = (rbuffersize / 2) - _size();
	while (toread > 0) {
	    int readnow = toread; 
	    int wrap_at = rbuffersize - rdmaoff; 
	    if (readnow > wrap_at)
		readnow = wrap_at;
	    int nread = read(fd, rbuf + rdmaoff, readnow);
	    if (nread < 0 && errno == EWOULDBLOCK)
		nread = 0;
	    assert (nread >= 0);
	    rdmaoff += nread;
	    if (rdmaoff == rbuffersize)
		rdmaoff = 0;
	    
	    toread -= nread;
	    // No use trying to read more if this read was short.
	    if (nread < readnow)
		break;
	}

	int sizebytes = _size();
	if (sizebytes > (rbuffersize * 3 / 4)) 
	    overruns++;

	int sizesamples = sizebytes / 2;
	return sizesamples;
    }

    s16 operator()() {
	assert(_size() >= 2);
	s16 s;
	s.v = (unsigned char) (rbuf[roff+0]) + 
	    ((signed char) (rbuf[roff+1]) << 8);
	roff += 2;
	if (roff == rbuffersize)
	    roff = 0;
	return s;
    }

    // Get an old sample.
    float operator[](int i) {
	int off = roff - 2*i;
	if (off < 0)
	    off += rbuffersize;
	int s =
	    (unsigned char) (rbuf[off+0]) + 
	    ((signed char) (rbuf[off+1]) << 8);
	return tofloat(s);
    }

    // Set the recording gain.  The argument goes from 0 to N_GAIN_STEPS-1.

    void set_gain(int gain_index) {
	int vol = gain_index * 99 / (N_GAIN_STEPS-1);
	input_gain = vol;
	if (ioctl(fd, SOUND_MIXER_WRITE_RECLEV, &vol) == -1) {
	    fprintf(stderr, "setting record level to %d: %s\n", 
		    vol, strerror(errno));
	    exit(1);
	}
    }

    // Reset the recording gain to a reasonable default value.

    void reset_gain() {
	set_gain(N_GAIN_STEPS * 2 / 3);
    }

private:    
    void mixer_ioctl(int ioctl_tag, char *desc, int value) {
	if (ioctl(fd, ioctl_tag, &value) == -1) {
	    fprintf(stderr, "setting mixer %s: %s\n", desc, strerror(errno));
	    //exit(1);
	}	
    }
    
    void flip_trigger(int bit) {
	int trigger;
	ioctl(fd, SNDCTL_DSP_GETTRIGGER, &trigger);
	trigger &= ~bit;
	ioctl(fd, SNDCTL_DSP_SETTRIGGER, &trigger);
	trigger |= bit;
	ioctl(fd, SNDCTL_DSP_SETTRIGGER, &trigger);
    }
      
//private:
public:
    int fd;
    char pbuf[BS];
    char rbuf[BS];
    int roff, poff;
    int rbuffersize, pbuffersize;
    int rdmaoff, pdmaoff;
    float input_gain;
    float output_gain;
  
    int overruns;
    int underruns;
};

#endif
