// $Id: linux-audio.h,v 1.12 2001/07/16 00:43:31 gson Exp $
//
// Linux/OSS audio driver for AuDSL.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

// This has only been tested with es1370-based cards and is unlikely to work
// with other ones yet.

#ifndef _LINUX_AUDIO_H
#define _LINUX_AUDIO_H

#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/mman.h>

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "dev-audio.h"

class LinuxAudioIO: public DevAudio {
public:
    // Initialize the audio device, configure it 
    // for full duplex operation, set an appropriate
    // (small) block size, etc.
    
    LinuxAudioIO(int audio_fd)
	: fd(audio_fd)
    {
	// Set the block size (a.k.a. fragment size)
	// 7FFF = no limit on number of blocks
	// 0008 = block size 2^8 
	int fragspec = 0x7FFF0008;
	if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &fragspec) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_SETFRAGMENT: %s\n", strerror(errno));
	    exit(1);
	}

	int format = AFMT_S16_LE;  
	if (ioctl(fd, SNDCTL_DSP_SETFMT, &format) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_SETFMT: %s\n", strerror(errno));
	    exit(1);
	}
	if (format != AFMT_S16_LE) {
	    fprintf(stderr, "audio device does not support the "
		    "S16_LE audio format\n");
	    exit(1);
	}
	    
	int stereo = 0;
	if (ioctl(fd, SNDCTL_DSP_STEREO, &stereo) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_STEREO: %s\n", strerror(errno));
	    exit(1);
	}
	if (stereo != 0) {
	    fprintf(stderr, "Audio device does not support mono mode\n");
	    exit(1);	    
	}

	int wanted_speed = 48000;
	int speed = wanted_speed;
	if (ioctl(fd, SNDCTL_DSP_SPEED, &speed) == -1)  {
	    fprintf(stderr, "SNDCTL_DSP_SPEED: %s\n", strerror(errno));
	    exit(1);
	}
	if (speed > wanted_speed * 11 / 10 || speed < wanted_speed * 9 / 10) {
	    fprintf(stderr, "Audio device does not support a "  
		    "sampling rate of %d Hz (closest %d Hz)\n",
		    wanted_speed, speed);
	    exit(1);
	}

	// Set suitable gains for the signal sources we actually use
	mixer_ioctl(SOUND_MIXER_WRITE_PCM, "DAC gain", 90);
	mixer_ioctl(SOUND_MIXER_WRITE_VOLUME, "master volume", 90);
	mixer_ioctl(SOUND_MIXER_WRITE_LINE, "LINE IN gain", 90);

	// Turn everything else off to avoid picking up unnecesary noise
	mixer_ioctl(SOUND_MIXER_WRITE_SYNTH, "SYNTH gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_CD, "CD gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_LINE1, "LINE1 gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_LINE2, "LINE2 gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_LINE3, "LINE3 gain", 0);
	mixer_ioctl(SOUND_MIXER_WRITE_MIC, "MIC gain", 0);

	// Record from LINE IN only
	mixer_ioctl(SOUND_MIXER_WRITE_RECSRC, "record source", 1 << SOUND_MIXER_LINE);
	
	int recsources = 1 << SOUND_MIXER_LINE;
	if (ioctl(fd, SOUND_MIXER_WRITE_RECSRC, &recsources) == -1) {
	    fprintf(stderr, "SOUND_MIXER_WRITE_RECSRC: %s\n", strerror(errno));
	    exit(1);
	}

	// Find the buffer size so that we can mmap it correctly
	struct audio_buf_info info;
	if (ioctl(fd, SNDCTL_DSP_GETOSPACE, &info) == -1) {
	    fprintf(stderr, "SNDCTL_DSP_GETOSPACE: %s\n", strerror(errno));
	    exit(1);
	}

	pbuffersize = rbuffersize = info.fragstotal * info.fragsize;
	
	roff = poff = 0;

	pbuf = (char *) mmap(NULL, pbuffersize, PROT_WRITE,
			     MAP_FILE|MAP_SHARED, fd, 0);
        if (pbuf == (char *) -1) {
	    fprintf(stderr, "playback mmap: %s\n", strerror(errno));
	    exit(1);
	}

	rbuf = (char *) mmap(NULL, rbuffersize, PROT_READ,
			      MAP_FILE|MAP_SHARED, fd, 0);
	if (rbuf == (char *) -1) {
	    fprintf(stderr, "record mmap: %s\n", strerror(errno));
	    exit(1);
	}
    }

    // Start recording, if your system does not
    // start it automatically when the audio device
    // is opened.
    
    void start_input() {
        flip_trigger(PCM_ENABLE_OUTPUT);
    }

    // Start playback, if your system does not
    // start it automatically when the audio device
    // is opened.
    
    void start_output() {
        flip_trigger(PCM_ENABLE_INPUT);      
    }

    // Get the number of samples currently in the playback buffer
    
    int playsize() {
	count_info info;
	int r = ioctl(fd, SNDCTL_DSP_GETOPTR, &info);
	if (r < 0) {
	    fprintf(stderr, "ioctl SNDCTL_DSP_GETOPTR: %s\n",
		    strerror(errno));
	    exit(1);
	}
	int sizebytes = poff - info.ptr;
	if (sizebytes <= 0)
	    sizebytes += pbuffersize;
	// If more than half the buffer, we have a probable underrun.
	if (sizebytes > (pbuffersize * 3 / 4)) {
	    underruns++;
	}
	int sizesamples = sizebytes / 2;
	return sizesamples;
    }

    // Get the number of samples currently in the recording buffer
    int size() {
	count_info info;
	int r = ioctl(fd, SNDCTL_DSP_GETIPTR, &info);
	if (r < 0) {
	    fprintf(stderr, "ioctl SNDCTL_DSP_GETIPTR: %s\n",
		    strerror(errno));
	    exit(1);
	}
	int sizebytes = info.ptr - roff;
	if (sizebytes < 0)
	    sizebytes += rbuffersize;
		
	if (sizebytes > (rbuffersize * 3 / 4)) {
	    overruns++;
	}

	int sizesamples = sizebytes / 2;
	
	// fprintf(stderr, "< s=%d d=%d o=%d, size %d samples\n",
	// ioffs.samples, ioffs.deltablks,
	// ioffs.offset, sizesamples);
	return sizesamples;
    }

    // Set the recording gain.  The argument goes from 0 to N_GAIN_STEPS-1.

    void set_gain(int gain_index) {
	int vol = gain_index * 99 / (N_GAIN_STEPS-1);
	input_gain = vol;
	if (ioctl(fd, SOUND_MIXER_WRITE_LINE, &vol) == -1) {
	    fprintf(stderr, "setting record level to %d: %s\n", 
		    vol, strerror(errno));
	    exit(1);
        }
    }

    // Reset the recording gain to a reasonable default value.

    void reset_gain() {
	set_gain(N_GAIN_STEPS * 2 / 3);
    }

private:    
    void mixer_ioctl(int ioctl_tag, char *desc, int value) {
	if (ioctl(fd, ioctl_tag, &value) == -1) {
	    fprintf(stderr, "setting mixer %s: %s\n", desc, strerror(errno));
	    exit(1);
        }	
    }
    
    void flip_trigger(int bit) {
        int trigger;
	ioctl(fd, SNDCTL_DSP_GETTRIGGER, &trigger);
	trigger &= ~bit;
	ioctl(fd, SNDCTL_DSP_SETTRIGGER, &trigger);
	trigger |= bit;
	ioctl(fd, SNDCTL_DSP_SETTRIGGER, &trigger);
    }
      

//private:
public:
    int fd;	// File descriptor for the audio device
};

#endif
