// $Id: rw-audio.h,v 1.16 2002/07/09 14:23:16 gson Exp $
//
// Read-write (non-memory-mapped) audio driver for NetBSD.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include "netbsd-audio.h"

// Full duplex audio through file descriptor read/write interface.
// Works with sockets as well as with the NetBSD /dev/audio.

#define BS 65536 // Ring buffer size

inline void printtime() {
    struct timeval tv;
    gettimeofday(&tv, 0);
    fprintf(stderr, "%lu.%06lu ", tv.tv_sec, tv.tv_usec);
}

class NetBSDRWAudioIO: public NetBSDAudio {
public:
    NetBSDRWAudioIO(int audio_fd)
	: NetBSDAudio(audio_fd)
    {
	roff = poff = 0;
	rdmaoff = pdmaoff = 0;
	rbuffersize = pbuffersize = BS;
	rbuf = new char[rbuffersize];
	pbuf = new char[pbuffersize];

	int r = fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) | O_NONBLOCK);
	if (r < 0) {
	    fprintf(stderr, "fcntl O_NONBLOCK: %s\n", strerror(errno));
	    exit(1);
	}

	while (read(fd, rbuf, rbuffersize) >= 1)
	    ;
    }

    ~NetBSDRWAudioIO() {
	delete [] rbuf;
	delete [] pbuf;
    }

    // Get the number of bytes of unplayed data in the play queue
    int _playsize() {
	int sizebytes = poff - pdmaoff;
	if (sizebytes < 0)
	    sizebytes += pbuffersize;
	return sizebytes;
    }

    // Get the number of unplayed samples in the play queue.
    int playsize() {
	int sizebytes = _playsize();
	if (sizebytes > (pbuffersize * 3 / 4)) {
	    underruns++;
	}

	int sizesamples = sizebytes / 2;
	return sizesamples;
    }
    
    void push(s16 s) {
	assert(_playsize() < pbuffersize - 4);

	DevAudio::push(s);

	// If we have a good-sized chunk of unplayed data, (try to) play it.
	int toplay = _playsize();

	if (toplay >= 128) {
	    while (toplay > 0) {
		// Number of bytes to play this iteration
		int playnow = toplay; 
		// Number of bytes before buffer wraps
		int wrap_at = pbuffersize - pdmaoff; 
		if (playnow > wrap_at)
		    playnow = wrap_at;
		int nplayed = write(fd, pbuf + pdmaoff, playnow);
		if (nplayed < 0 && errno == EWOULDBLOCK)
		    nplayed = 0;
		assert (nplayed >= 0);
		pdmaoff += nplayed;
		if (pdmaoff == pbuffersize)
		    pdmaoff = 0;
		toplay -= nplayed;
		// No use trying to play more if this play was short.
		if (nplayed < playnow)
		    break;
	    }
	}
    }

    int _size() {
	int sizebytes = rdmaoff - roff;
	if (sizebytes < 0)
	    sizebytes += rbuffersize;
	return sizebytes;
    }

    int size() {
	// First try to fill the buffer by doing nonblocking
	// reads from our file descriptor.
	// Try to make the buffer half full.
	int toread = (rbuffersize / 2) - _size();
	while (toread > 0) {
	    int readnow = toread; 
	    int wrap_at = rbuffersize - rdmaoff; 
	    if (readnow > wrap_at)
		readnow = wrap_at;
	    int nread = read(fd, rbuf + rdmaoff, readnow);
	    if (nread < 0 && errno == EWOULDBLOCK) {
		nread = 0;
	    }
	    assert (nread >= 0);
	    rdmaoff += nread;
	    if (rdmaoff == rbuffersize)
		rdmaoff = 0;
	    
	    toread -= nread;
	    // No use trying to read more if this read was short.
	    if (nread < readnow)
		break;
	}

	int sizebytes = _size();
	if (sizebytes > (rbuffersize * 3 / 4)) {
	    overruns++;
	}

	int sizesamples = sizebytes / 2;
	return sizesamples;
    }

    int rdmaoff; // Offset in record buffer where kerenel data is written
    int pdmaoff; // Offset in play buffer where kernel data is read
};
