// $Id: fdtunnel.cc,v 1.3 2004/11/30 14:53:00 gson Exp $
//
// Full duplex audio I/O for use in pipelines.
//
// (C) Copyright 1999 Araneus Information Systems Oy.  All rights reserved.
// Please refer to the file COPYRIGHT for detailed copyright information.
//

// Simultaneously plays audio data from stdin and records onto stdout.
//
// Example: sine_generator | fdpipe | spectrum_analyzer.

#include <stdio.h>

#include "open-audio.h"
#include "audio-io.h"
#include "fileio.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

// #define BSIZ 4096
#define BSIZ 256

char *progname;

// Log an error message and exit

void error(char *s) {
    fprintf(stderr, "%s: %s: %s\n", progname, s, strerror(errno));
    exit(1);
}

// This structure represents a single (unidirectional) copy 
// operation in progress, like a single instance of the "cat"
// command.
  
typedef struct {
    char buf[BSIZ];
    int r;	/* read file descriptor */
    int w;	/* write file descriptor */
    char *p;	/* current position in buffer */
    int n;	/* amount of unwritten data in buffer */
    int eof;	/* 0 if open; 1 if eof received; 2 if eof sent */
    int copied;
} cat;

// Do N simultaneous "cat" operations; return when they have all
// seen EOF.

void copy(int catc, cat *catv)
{
    fd_set readfds, writefds, exceptfds;
    int i, n;
    
    for (i = 0; i < catc; i++) {
	cat *c = &catv[i];
	c->n = 0;
	c->eof = 0;
	c->copied = 0;
    }
    
    for (;;) {
	int max;

	FD_ZERO(&readfds); FD_ZERO(&writefds); FD_ZERO(&exceptfds);
	max = -1;
	
	for (i = 0; i < catc; i++) {
	    cat *c = &catv[i];
	    if (c->n || c->eof == 1) {
		FD_SET(c->w, &writefds);
		if (c->w > max)
		    max = c->w;
	    } else if (!c->eof) {
		FD_SET(c->r, &readfds);
		if (c->r > max)
		    max = c->r;
	    }
	}

	// fprintf(stderr, "entering select, %d %d %d\n", catv[0].eof, catv[0].copied, catv[1].copied);

	if (catv[0].eof && catv[1].copied >= catv[0].copied)
	    break;
	    
	n = select(max+1, &readfds, &writefds, &exceptfds,
		   (struct timeval *) 0);
	if (n == -1)
	    error("select");

	for (i = 0; i < catc; i++) {
	    cat *c = &catv[i];
	    if (FD_ISSET(c->r, &readfds)) {
		n = read(c->r, c->buf, BSIZ);
		if (n == 0) {
		    fprintf(stderr, "eof reading fd %d\n", c->r);
		    c->eof = 1;
		} else if (n == -1) {
		    fprintf(stderr, "error reading cat %d\n", i);
		    error("read");
		} else {
		    c->n = n;
		    c->p = c->buf;
		}
	    }
	    if (FD_ISSET(c->w, &writefds)) {
		if (c->n) {
		    n = write(c->w, c->p, c->n);
		    if (n == -1)
			error("write");

		    c->n -= n;
		    c->p += n;
		    c->copied += n;
		}
	    }	    
	}
    }
}
    
int main(int argc, char **argv) {
    int blocksize = 480;
    progname = argv[0];

    argv++, argc--;
    if (argc > 1 && strcmp(argv[0], "--blocksize") == 0) {
	blocksize = atoi(argv[1]);
	argv += 2; argc -= 2;
    }
    if (argc != 1)
	error("usage: fdtunnel [--blocksize n] device");
			   
    int fd = open_audio(argv[0]);

    // Check audio device properties
    int props;
    int r = ioctl(fd, AUDIO_GETPROPS, &props);
    if ((props & AUDIO_PROP_FULLDUPLEX) != AUDIO_PROP_FULLDUPLEX) {
	fprintf(stderr, "Audio device does not support "
		"full duplex.  Sorry.\n");
	exit(1);
    }
    
    // Set full duplex mode
    int yes = 1;
    r = ioctl(fd, AUDIO_SETFD, &yes);
    if (r < 0) {
	fprintf(stderr, "ioctl AUDIO_SETFD: %s\n",
		strerror(errno));
	exit(1);
    }
    
    // Set an appropriate block size
    audio_info_t info;	
    AUDIO_INITINFO(&info);
    info.blocksize = blocksize;
    // Due to a bug in the audio driver, the block size may not actually get
    // set unless we set info.mode = AUMODE_RECORD | AUMODE_PLAY redundantly.
    info.mode = AUMODE_RECORD | AUMODE_PLAY;
    r = ioctl(fd, AUDIO_SETINFO, &info);
    if (r < 0) {
	fprintf(stderr, "ioctl AUDIO_SETINFO blocksize/mode: %s\n",
		strerror(errno));
	exit(1);
    }

    AUDIO_INITINFO(&info);
    info.record.pause = 0;
    r = ioctl(fd, AUDIO_SETINFO, &info);
    if (r < 0) {
	fprintf(stderr, "ioctl AUDIO_SETINFO unpause: %s\n",
		strerror(errno));
	exit(1);
    }
    
    cat cats[2];
    cats[0].r = 0;
    cats[0].w = fd;
    cats[1].r = fd;
    cats[1].w = 1;
    copy(2, cats);
    fprintf(stderr, "copied %d bytes\n", cats[1].copied);
    close(fd);
    return 0;
}
